import pytest
import os
import tempfile

from test_load_and_store import _load_oval_document


DATA_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "data"))
OVAL_DOCUMENT_PATH = os.path.join(
    DATA_DIR, "minimal_oval_of_oval_ssg-sshd_rekey_limit_def.xml"
)
TEST_BUILD_DIR = tempfile.mkdtemp()


@pytest.fixture
def oval_document():
    return _load_oval_document(OVAL_DOCUMENT_PATH)


def test_all_references_of_definition(oval_document):
    ref = oval_document.get_all_references_of_definition(
        "oval:ssg-sshd_not_required_or_unset:def:1"
    )
    assert ref.is_done()
    assert ref.definitions == [
        "oval:ssg-sshd_not_required_or_unset:def:1",
        "oval:ssg-sshd_requirement_unset:def:1",
    ]
    assert ref.tests == [
        "oval:ssg-test_sshd_not_required:tst:1",
        "oval:ssg-test_sshd_requirement_unset:tst:1",
    ]
    assert ref.objects == [
        "oval:ssg-object_sshd_requirement_unknown:obj:1",
        "oval:ssg-object_sshd_not_required:obj:1",
    ]
    assert ref.states == [
        "oval:ssg-state_sshd_requirement_unset:ste:1",
        "oval:ssg-state_sshd_not_required:ste:1",
    ]
    assert ref.variables == [
        "oval:ssg-sshd_required:var:1",
    ]


def _assert_filtered_oval_document(oval_document_, def_id):
    assert def_id in oval_document_.definitions
    assert "oval:ssg-sshd_rekey_limit:def:1" not in oval_document_.definitions
    assert "oval:ssg-test_sshd_not_required:tst:1" in oval_document_.tests
    assert "oval:ssg-test_sshd_rekey_limit:tst:1" not in oval_document_.tests
    assert "oval:ssg-object_sshd_requirement_unknown:obj:1" in oval_document_.objects
    assert "oval:ssg-obj_sshd_rekey_limit:obj:1" not in oval_document_.objects
    assert "oval:ssg-state_sshd_requirement_unset:ste:1" in oval_document_.states
    assert "oval:ssg-state_sshd_rekey_limit:ste:1" not in oval_document_.states
    assert "oval:ssg-sshd_required:var:1" in oval_document_.variables
    assert "oval:ssg-var_rekey_limit_size:var:1" not in oval_document_.variables


def test_keep_referenced_components(oval_document):
    def_id = "oval:ssg-sshd_not_required_or_unset:def:1"
    ref = oval_document.get_all_references_of_definition(def_id)
    oval_document.keep_referenced_components(ref)
    _assert_filtered_oval_document(oval_document, def_id)


@pytest.mark.parametrize(
    "path, expected_result",
    [
        (
           OVAL_DOCUMENT_PATH, True,
        ),
        (
            os.path.join(DATA_DIR, "oval_with_broken_extend_definition.xml"),
            False
        ),
        (
            os.path.join(DATA_DIR, "oval_with_correct_extend_definition.xml"),
            True
        )
    ]
)
def test_validation(path, expected_result):
    oval_doc = _load_oval_document(path)
    assert oval_doc.validate_references() == expected_result


def test_save_as_xml(oval_document):
    oval_doc_path = os.path.join(TEST_BUILD_DIR, "oval.xml")
    def_id = "oval:ssg-sshd_not_required_or_unset:def:1"

    with open(oval_doc_path, "wb") as fd:
        refs = oval_document.get_all_references_of_definition(def_id)
        oval_document.save_as_xml(fd, refs)

    filtered_oval_document = _load_oval_document(oval_doc_path)
    _assert_filtered_oval_document(filtered_oval_document, def_id)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import yaml

from .controls import ControlsManager, Policy
from .products import (
    get_profile_files_from_root,
    load_product_yaml,
    product_yaml_path,
)


if sys.version_info >= (3, 9):
    dict_type = dict    # Python 3.9+ supports built-in generics
    list_type = list
    tuple_type = tuple
else:
    from typing import Dict as dict_type    # Fallback for older versions
    from typing import List as list_type
    from typing import Tuple as tuple_type


class ProfileSelections:
    """
    A class to represent profile with sections of rules and variables.

    Attributes:
    -----------
    profile_id : str
        The unique identifier for the profile.
    profile_title : str
        The profile title associated with the profile id.
    product_id : str
        The product id associated with the profile.
    product_title : str
        The product title associated with the product id.
    """
    def __init__(self, profile_id, profile_title, product_id, product_title):
        self.profile_id = profile_id
        self.profile_title = profile_title
        self.product_id = product_id
        self.product_title = product_title
        self.rules = []
        self.unselected_rules = []
        self.variables = {}


def _load_product_yaml(content_dir: str, product: str) -> object:
    """
    Load the product YAML file and return its content as a Python object.

    Args:
        content_dir (str): The directory where the content is stored.
        product (str): The name of the product.

    Returns:
        object: The loaded YAML content as a Python object.
    """
    file_yaml_path = product_yaml_path(content_dir, product)
    return load_product_yaml(file_yaml_path)


def _load_yaml_profile_file(file_path: str) -> dict_type:
    """
    Load the content of a YAML file intended to profiles definitions.

    It is not necessary to process macros in this case.

    Args:
        file_path (str): The path to the YAML file.

    Returns:
        dict: The content of the YAML file as a dictionary.
    """
    with open(file_path, 'r') as file:
        try:
            return yaml.safe_load(file)
        except yaml.YAMLError as e:
            print(f"Error loading YAML profile file {file_path}: {e}")
            return {}


def _get_extended_profile_path(profiles_files: list, profile_name: str) -> str:
    """
    Retrieve the full path of a profile file from a list of profile file paths.

    Args:
        profiles_files (list of str): A list of file paths where profile files are located.
        profile_name (str): The name of the profile to search for.

    Returns:
        str: The full path of the profile file if found, otherwise None.
    """
    profile_file = f"{profile_name}.profile"
    profile_path = next((path for path in profiles_files if profile_file in path), None)
    return profile_path


def _process_profile_extension(profile: ProfileSelections, profile_yaml: dict,
                               profiles_files: list, policies: dict) -> ProfileSelections:
    """
    Processes the extension of a profile by recursively checking if the profile extends another
    profile and updating the profile selections accordingly.

    Args:
        profile (ProfileSelections): The profile object to be processed.
        profile_yaml (dict): The YAML content of the current profile.
        profiles_files (list): List of profile file paths.
        policies (dict): The policies defined in the current profile.

    Returns:
        ProfileSelections: The updated profile object.
    """
    extended_profile = profile_yaml.get("extends")
    if isinstance(extended_profile, str):
        extended_profile = _get_extended_profile_path(profiles_files, extended_profile)
        if extended_profile is not None:
            profile_yaml = _load_yaml_profile_file(extended_profile)
            return _process_profile(profile, profile_yaml, profiles_files, policies)
    return profile


def _parse_control_line(control_line: str) -> tuple_type[str, str]:
    """
    Parses a control line string and returns a tuple containing the first and third parts of the
    string, separated by a colon. If the string does not contain three parts, the second element
    of the tuple defaults to 'all'.

    Args:
        control_line (str): The control line string to be parsed.

    Returns:
        tuple[str, str]: A tuple containing the first part of the control line and either the
                         third part or 'all' if the third part is not present.
    """
    parts = control_line.split(":")
    if len(parts) == 3:
        return parts[0], parts[2]
    return parts[0], 'all'


def _process_selected_variable(profile: ProfileSelections, variable: str) -> None:
    """
    Processes a selected variable and updates the profile's variables.

    Args:
        profile (ProfileSelections): The profile object containing variables.
        variable (str): The variable in the format 'name=value'.

    Raises:
        ValueError: If the variable is not in the correct format.
    """
    variable_name, variable_value = variable.split('=', 1)
    if variable_name not in profile.variables:
        profile.variables[variable_name] = variable_value


def _process_selected_rule(profile: ProfileSelections, rule: str) -> None:
    """
    Adds a rule to the profile's selected rules if it is not already selected or unselected.

    Args:
        profile (ProfileSelections): The profile containing selected and unselected rules.
        rule (str): The rule to be added to the profile's selected rules.

    Returns:
        None
    """
    if rule not in profile.unselected_rules and rule not in profile.rules:
        profile.rules.append(rule)


def _process_control(profile: ProfileSelections, control: object) -> None:
    """
    Processes a control by iterating through its rules and applying the appropriate processing
    function. Not that at this level rules list in control can include both variables and rules.
    The function distinguishes between variable and rules based on the presence of an '='
    character in the rule.

    Args:
        profile (ProfileSelections): The profile selections to be processed.
        control: The control object containing rules to be processed.
    """
    for rule in control.rules:
        if "=" in rule:
            _process_selected_variable(profile, rule)
        else:
            _process_selected_rule(profile, rule)


def _update_profile_with_policy(profile: ProfileSelections, policy: Policy, level: str) -> None:
    """
    Updates the given profile with controls from the specified policy based on the provided level.

    Args:
        profile (ProfileSelections): The profile to be updated.
        policy (Policy): The policy containing controls to update the profile with.
        level (str): The level of controls to be processed. If 'all', all controls are processed.
                     Otherwise, only controls matching the specified level are processed.

    Returns:
        None
    """
    for control in policy.controls:
        if level == 'all' or level in control.levels:
            _process_control(profile, control)


def _process_controls(profile: ProfileSelections, control_line: str,
                      policies: dict) -> ProfileSelections:
    """
    Process a control file inheritance to update profile selections based on the given policies.

    Args:
        profile (ProfileSelections): The profile object to be processed.
        control_line (str): A string representing the control line, which contains a policy ID and
                            optionally a level, separated by colons.
        policies (dict): A dictionary of policies, where each key is a policy ID and each value is
                         a policy object containing the controls.

    Returns:
        ProfileSelections: The updated profile object.
    """
    policy_id, level = _parse_control_line(control_line)
    policy = policies.get(policy_id)

    if policy is None:
        print(f"Policy {policy_id} not found")
        return profile

    _update_profile_with_policy(profile, policy, level)
    return profile


def _process_selections(profile: ProfileSelections, profile_yaml: dict,
                        policies: dict) -> ProfileSelections:
    """
    Processes the selections from the profile YAML and updates the profile accordingly.

    Args:
        profile (ProfileSelections): The profile object to be processed.
        profile_yaml (dict): A dictionary containing the profile YAML data.
        policies (dict): A dictionary containing policy information.

    Returns:
        profile: The updated profile object.
    """
    selections = profile_yaml.get("selections", [])
    for selected in selections:
        if selected.startswith("!"):
            profile.unselected_rules.append(selected[1:])
        elif "=" in selected:
            variable_name, variable_value = selected.split('=', 1)
            profile.variables[variable_name] = variable_value
        elif ":" in selected:
            profile = _process_controls(profile, selected, policies)
        else:
            profile.rules.append(selected)
    return profile


def _process_profile(profile: ProfileSelections, profile_yaml: dict, profiles_files: list,
                     policies: dict) -> ProfileSelections:
    """
    Processes a profile by handling profile extensions, and processing selections.

    Args:
        profile (ProfileSelections): The profile object to be processed.
        profile_yaml (dict): The YAML content of the profile.
        profiles_files (list): A list of profile file paths.
        policies (dict): A dictionary of policies defined by control files.

    Returns:
        ProfileSelections: The processed profile object.
    """
    profile = _process_profile_extension(profile, profile_yaml, profiles_files, policies)
    profile = _process_selections(profile, profile_yaml, policies)
    return profile


def _load_controls_manager(controls_dir: str, product_yaml: dict) -> object:
    """
    Loads and initializes a ControlsManager instance.

    Args:
        controls_dir (str): The directory containing control files.
        product_yaml (dict): The product configuration in YAML format.

    Returns:
        object: An instance of ControlsManager with loaded controls.
    """
    control_mgr = ControlsManager(controls_dir, product_yaml)
    control_mgr.load()
    return control_mgr


def _sort_profiles_selections(profiles: list) -> ProfileSelections:
    """
    Sorts profiles selections (rules and variables) by selections ids.

    Args:
        profiles (list): A list of ProfileSelections objects to be sorted.

    Returns:
        ProfileSelections: The sorted list of ProfileSelections objects.
    """
    for profile in profiles:
        profile.rules = sorted(profile.rules)
        profile.unselected_rules = sorted(profile.unselected_rules)
        profile.variables = dict(sorted(profile.variables.items()))
    return profiles


def get_profiles_from_products(content_dir: str, products: list,
                               sorted: bool = False) -> list_type:
    """
    Retrieves profiles with respective variables from the given products.

    Args:
        content_dir (str): The directory containing the content.
        products (list): A list of product names to retrieve profiles from.

    Returns:
        list: A list of ProfileVariables objects containing profile variables for each product.
    """
    profiles = []
    controls_dir = os.path.join(content_dir, 'controls')

    for product in products:
        product_yaml = _load_product_yaml(content_dir, product)
        product_title = product_yaml.get("full_name")
        profiles_files = get_profile_files_from_root(product_yaml, product_yaml)
        controls_manager = _load_controls_manager(controls_dir, product_yaml)
        for file in profiles_files:
            profile_id = os.path.basename(file).split('.profile')[0]
            profile_yaml = _load_yaml_profile_file(file)
            profile_title = profile_yaml.get("title")
            profile = ProfileSelections(profile_id, profile_title, product, product_title)
            profile = _process_profile(profile, profile_yaml, profiles_files,
                                       controls_manager.policies)
            profiles.append(profile)

    if sorted:
        profiles = _sort_profiles_selections(profiles)

    return profiles



#!/usr/bin/python3

import argparse
import subprocess
import sys
import textwrap
import os
import re
import time
import yaml

from ssg.utils import mkdir_p

class JinjaString(str):
    pass

def jinja_string_representer(dumper, data):
    # yuumasato: The Jinja syntax is not compatible with YAML as curly braces are special characters.
    # So we use an exotic token to mark where opening and closing Jinja tags would be, they are substituted
    # by actual `{{{` and `}}}` right before being written to file.
    # A the moment this is restricted to macro calls, but could be expanded to conditionals as well.
    sanitized = re.sub(r"{{{", "JiNjA_OpEn", data)
    sanitized = re.sub(r"}}}", "JiNjA_ClOsE", sanitized)
    return dumper.represent_scalar(u'tag:yaml.org,2002:str', sanitized, style="|")

yaml.add_representer(JinjaString, jinja_string_representer)

PROG_DESC = (''' Create and test content files for Kubernetes API checks.

This script is intended to help content writers create a new application check
for OCP4/Kubernetes.

- The 'create' subcommand creates the initial files for a new rule and fetches
  the raw URL of the object in question (unless you specify the URL).

- The 'test' subcommand builds your content locally and tests directly using an
  openscap podman container. The scan container will test against yaml files
  staged under --objectdir.

- The 'cluster-test' subcommand pushes the content to your cluster, and then
  runs a Platform scan for your rule with compliance-operator.

Example workflow:

$ utils/add_kubernetes_rule.py create --rule=ocp_proxy_has_ca \
  --type="proxies.config" --name="cluster" \
  --yamlpath=".spec.trustedCA.name" --match="[a-zA-Z0-9]*"
creating check for "/apis/config.openshift.io/v1/proxies/cluster" with yamlpath ".spec.trustedCA.name" satisfying match of "[a-zA-Z0-9]*"
wrote applications/openshift/ocp_proxy_has_ca/rule.yml

$ mkdir -p /tmp/apis/config.openshift.io/v1/proxies/
$ oc get proxies.config/cluster -o yaml > /tmp/apis/config.openshift.io/v1/proxies/cluster
$ utils/add_kubernetes_rule.py test --rule=ocp_proxy_has_ca
testing rule ocp_proxy_has_ca locally
Title
        None
Rule
        xccdf_org.ssgproject.content_rule_ocp_proxy_has_ca
Ident
        CCE-84209-6
Result
        pass

$ utils/add_kubernetes_rule.py cluster-test --rule=ocp_proxy_has_ca
testing rule ocp_proxy_has_ca in-cluster
deploying compliance-operator
pushing image build to cluster
waiting for cleanup from previous test run
output from last phase check: LAUNCHING NOT-AVAILABLE
output from last phase check: RUNNING NOT-AVAILABLE
output from last phase check: AGGREGATING NOT-AVAILABLE
output from last phase check: DONE COMPLIANT
COMPLIANT

''')

OCP_RULE_DIR = 'applications/openshift'
OSCAP_TEST_IMAGE = 'quay.io/compliance-operator/openscap-ocp:1.3.4'
OSCAP_CMD_TEMPLATE = 'oscap xccdf eval --verbose %s --fetch-remote-resources --profile xccdf_org.ssgproject.content_profile_test --results-arf /tmp/report-arf.xml /content/ssg-ocp4-ds.xml'
PROFILE_PATH = 'products/ocp4/profiles/test.profile'

MOCK_VERSION = ('''status:
  versions:
  - name: operator
    version: 4.6.0-0.ci-2020-06-15-112708
  - name: openshift-apiserver
    version: 4.6.0-0.ci-2020-06-15-112708
''')


def set_operation_value(value, template_vars):
    if value:
        template_vars['operation'] = 'pattern match'
        template_vars['type'] = 'string'

def set_entity_value(value, template_vars):
    if value is not None:
        template_vars['entity_check'] = value

def set_check_existence_value(value, template_vars):
    if value is not None:
        template_vars['check_existence'] = value

def set_template_vars(value, template_vars):
    for var in value.split(","):
        key, value = var.strip().split(":")
        template_vars[key.strip()] = value.strip()

PROFILE_TEMPLATE = ('''documentation_complete: true

title: 'Test Profile for {RULE_NAME}'

platform: ocp4

description: Test Profile
selections:
- {RULE_NAME}
''')


TEST_SCAN_TEMPLATE = ('''apiVersion: compliance.openshift.io/v1alpha1
kind: ComplianceScan
metadata:
  name: test
  namespace: {NAMESPACE}
spec:
  scanType: {TYPE}
  profile: {PROFILE}
  content: ssg-ocp4-ds.xml
  contentImage: image-registry.openshift-image-registry.svc:5000/openshift-compliance/openscap-ocp4-ds:latest
  debug: true
''')


def needs_oc(func):
    def wrapper(args):
        if which('oc') is None:
            print('oc is required for this command.')
            return 1

        return func(args)
    return wrapper


def needs_working_cluster(func):
    def wrapper(args):
        ret_code, output = subprocess.getstatusoutput(
            'oc whoami')
        if ret_code != 0:
            print("* Error connecting to cluster")
            print(output)
            return ret_code

        return func(args)
    return wrapper

def which(program):
    fpath, fname = os.path.split(program)
    if fpath:
        if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
                return exe_file

    return None


def create_base_rule(args, url=None, node_rule=False):
    rule_yaml = dict()
    rule_yaml['documentation_complete'] = True
    rule_yaml['title'] = args.title
    if node_rule:
        rule_yaml['platform'] = 'ocp4-node'
    rule_yaml['description'] = args.description
    rule_yaml['rationale'] = 'TBD'
    rule_yaml['identifiers'] = dict()
    rule_yaml['severity'] = args.severity
    if args.jqfilter:
        rule_yaml['warnings'] = [{'general': JinjaString("{{{ openshift_filtered_cluster_setting({'%s': '%s'}) | indent(4) }}}" % (url, args.jqfilter))}]
    elif url:
        rule_yaml['warnings'] = [{'general': JinjaString('{{{ openshift_cluster_setting("%s") | indent(4) }}}' % (url))}]
    rule_yaml['template'] = dict()

    return rule_yaml


def save_rule(rule_yaml_path, rule_yaml):
    with open(rule_yaml_path, 'w') as f:
        yaml_contents = yaml.dump(rule_yaml, None, indent=4, sort_keys=False, canonical=False, default_flow_style=False, width=120)
        # Adds a blank line between keys
        formatted_yaml_contents = re.sub(r"\n(\w+:.*)", r"\n\n\1", yaml_contents)

        # Replace placeholders for CaC/content Jinja2 expressions
        formatted_yaml_contents = re.sub(r"JiNjA_OpEn", r"{{{", formatted_yaml_contents)
        formatted_yaml_contents = re.sub(r"JiNjA_ClOsE", r"}}}", formatted_yaml_contents)

        f.write(formatted_yaml_contents)
    print('* Wrote ' + rule_yaml_path)


def createNodeRuleFunc(args):
    group_path = os.path.join(OCP_RULE_DIR, args.group)
    if args.group:
        if not os.path.isdir(group_path):
            print("ERROR: The specified group '%s' doesn't exist in the '%s' directory" % (
                args.group, OCP_RULE_DIR))
            return 0

    rule_path = os.path.join(group_path, args.rule)
    rule_yaml_path = os.path.join(rule_path, 'rule.yml')

    mkdir_p(rule_path)

    rule_yaml = create_base_rule(args)

    template = rule_yaml['template']
    template['name'] = args.template

    template['vars'] = dict()
    template_vars = set_template_vars(args.template_vars, template['vars'])

    save_rule(rule_yaml_path, rule_yaml)
    return 0


@needs_oc
def createPlatformRuleFunc(args):
    url = args.url
    retries = 0
    namespace_flag = ''
    if args.namespace is not None:
        namespace_flag = '-n ' + args.namespace
    elif args.all_namespaces:
        namespace_flag = '-A'

    group_path = os.path.join(OCP_RULE_DIR, args.group)
    if args.group:
        if not os.path.isdir(group_path):
            print("ERROR: The specified group '%s' doesn't exist in the '%s' directory" % (
                args.group, OCP_RULE_DIR))
            return 0

    rule_path = os.path.join(group_path, args.rule)
    while url is None and retries < 5:
        retries += 1
        cmdstr = 'oc get %s' % (args.type)

        if args.name:
            cmdstr += ' ' + args.name

        cmdstr += ' %s --loglevel=6' % (namespace_flag)

        print("Running: " + cmdstr)
        ret_code, output = subprocess.getstatusoutput(cmdstr)

        if ret_code != 0:
            print('error running oc, check connection to the cluster: %d\n %s' % (
                ret_code, output))
            continue

        fetch_line = ""
        url_part = ""
        lines = output.splitlines()
        for line in lines:
            if 'GET' in line:
                fetch_line = line
                break

        if len(fetch_line) > 0:
            # extract the object url from the debug line
            full_url = fetch_line[fetch_line.index("GET"):].split(" ")[1]
            url_part = full_url[full_url.rfind("/api"):]

        if len(url_part) > 0 and '/api' in url_part:
            url = url_part

    if url is None:
        print('there was a problem finding the URL from the oc debug output. Hint: override this automatic check with --url')
        return 1

    print('* Creating check for "%s" with yamlpath "%s" satisfying match of "%s"' % (
        url, args.yamlpath, args.match))
    rule_yaml_path = os.path.join(rule_path, 'rule.yml')

    mkdir_p(rule_path)

    rule_yaml = create_base_rule(args, url)

    template = rule_yaml['template']
    template['name'] = 'yamlfile_value'

    template['vars'] = dict()
    template_vars = template['vars']
    template_vars['ocp_data'] = "true"
    if args.jqfilter:
        template_vars['filepath'] = JinjaString("{{{ openshift_filtered_path('%s', '%s') }}}" % (url, args.jqfilter))
    else:
        template_vars['filepath'] = url
    template_vars['yamlpath'] = args.yamlpath

    set_entity_value(args.match_entity, template_vars)
    set_check_existence_value(args.check_existence, template_vars)

    if args.match:
        value_dict = dict()
        value_dict['value'] = args.match
        set_operation_value(args.regex, value_dict)

        template_vars['values'] = [value_dict]
    else:
        template_vars['xccdf_variable'] = args.variable

    save_rule(rule_yaml_path, rule_yaml)

    return 0


def createTestProfile(rule):
    # create a solo profile for rule
    with open(PROFILE_PATH, 'w') as f:
        f.write(PROFILE_TEMPLATE.format(RULE_NAME=rule))


@needs_oc
@needs_working_cluster
def clusterTestFunc(args):

    print('* Testing rule %s in-cluster' % args.rule)

    findout = subprocess.getoutput(
        "find %s -name '%s' -type d" % (OCP_RULE_DIR, args.rule))
    if findout == "":
        print('ERROR: no rule for %s, run "create" first' % args.rule)
        return 1

    if not args.skip_deploy:
        subprocess.run("utils/deploy_compliance_operator.sh")

    if not args.skip_build:
        createTestProfile(args.rule)
        print('* Pushing image build to cluster')
        # execute the build_ds_container script
        buildp = subprocess.run(
            ['utils/build_ds_container.py', '-P', 'ocp4', 'rhcos4'])
        if buildp.returncode != 0:
            try:
                os.remove(PROFILE_PATH)
            except OSError:
                pass
            return 1

    ret_code, _ = subprocess.getstatusoutput(
        'oc delete -n {NAMESPACE} compliancescans/test'.format(NAMESPACE=args.namespace))
    if ret_code == 0:
        # if previous compliancescans were actually deleted, wait a bit to allow resources to clean up.
        print('* Waiting for cleanup from a previous test run')
        time.sleep(20)

    # create a single-rule scan
    print("* Running scan with rule '%s'" % args.rule)
    profile = 'xccdf_org.ssgproject.content_profile_test'
    apply_cmd = ['oc', 'apply', '-f', '-']
    with subprocess.Popen(apply_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE) as proc:
        _, err = proc.communicate(
            input=TEST_SCAN_TEMPLATE.format(PROFILE=profile, TYPE=args.scantype, NAMESPACE=args.namespace).encode())
        if proc.returncode != 0:
            print('Error applying scan object: %s' % err)
            try:
                os.remove(PROFILE_PATH)
            except OSError:
                pass
            return 1

    # poll for the DONE result
    timeout = time.time() + 120   # A couple of minutes is generous for the platform scan.
    scan_result = None
    while True:
        ret_code, output = subprocess.getstatusoutput(
            'oc get -n {NAMESPACE} compliancescans/test -o template="{{{{.status.phase}}}} {{{{.status.result}}}}"'.format(NAMESPACE=args.namespace))
        if output is not None:
            print('> Output from last phase check: %s' % output)
        if output.startswith('DONE'):
            scan_result = output[5:]
            break
        if time.time() >= timeout:
            break
        time.sleep(2)

    if scan_result is None:
        print('ERROR: Timeout waiting for scan to finish')
        return 1

    print("* The result is '%s'" % scan_result)
    return 0


def testFunc(args):
    if which('podman') is None:
        print('podman is required')
        return 1

    print('testing rule %s locally' % args.rule)

    if not args.skip_build:
        createTestProfile(args.rule)
        ret_code, out = subprocess.getstatusoutput('./build_product --datastream-only ocp4')
        if ret_code != 0:
            print('build failed: %s' % out)
            return 1

    # mock a passing result for the implicit ocp4 version check
    version_dir = args.objectdir + '/apis/config.openshift.io/v1/clusteroperators'
    mock_version_file = os.path.join(version_dir, 'openshift-apiserver')
    if not os.path.exists(mock_version_file):
        mkdir_p(version_dir)
        with open(mock_version_file, 'w') as f:
            f.write(MOCK_VERSION)

    oscap_cmd_opts = OSCAP_CMD_TEMPLATE % (args.verbosity)
    pod_cmd = 'podman run -it --security-opt label=disable -v "%s:/content" -v "%s:/kubernetes-api-resources" %s %s' % (args.contentdir,
                                                                                                                        args.objectdir, OSCAP_TEST_IMAGE, oscap_cmd_opts)
    print(subprocess.getoutput(pod_cmd))


def main():
    parser = argparse.ArgumentParser(
        prog="add_kubernetes_rule.py",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(PROG_DESC))
    subparser = parser.add_subparsers(
        dest='subcommand', title='subcommands', help='pick one')
    create_parser = subparser.add_parser(
        'create', help='Bootstrap the XML and YML files under %s for a new check.' % OCP_RULE_DIR)

    common_rule_args = argparse.ArgumentParser(add_help=False)
    common_rule_args.add_argument(
        '--rule', required=True, help='The name of the rule to create. Required.')
    common_rule_args.add_argument(
        '--group', default="", help='The group directory of the rule to create.')
    common_rule_args.add_argument(
        '--name', help='The name of the Kubernetes object to check.')
    common_rule_args.add_argument(
        '--title', help='A short description of the check.')
    common_rule_args.add_argument(
        '--description', help='A human-readable description of the provided matching criteria.')
    common_rule_args.add_argument(
        '--severity', default="unknown", help='the severity of the rule.')
    common_rule_args.add_argument(
        '--identifiers', default="TBD", help='an identifier for the rule (CCE number)')
    common_rule_args.add_argument(
        '--jqfilter', default="", help='A JQ filter to select the data passed down for OVAL evaluation.')

    type_parser = create_parser.add_subparsers(dest='rule types', title='Creates a rule', help='Types of rules')
    platform_parser = type_parser.add_parser('platform', help='Creates a Platform rule',  parents=[common_rule_args])
    platform_parser.add_argument('--yamlpath',
                               help='The yaml-path of the element to match against.')
    value_or_variable = platform_parser.add_mutually_exclusive_group()
    value_or_variable.add_argument(
        '--match', help='A string value or regex providing the matching criteria. One of "match" or "variable" are required')
    value_or_variable.add_argument(
        '--variable', help='A string name of the XCCDF variable to with the value to check for. Mutually exclusive with "match" option')
    platform_parser.add_argument(
        '--namespace', help='The namespace of the Kubernetes object (optional for cluster-scoped objects)', default=None)
    platform_parser.add_argument(
        '--all-namespaces', action="store_true", help='The namespace of the Kubernetes object (optional for cluster-scoped objects)',
        default=False)
    platform_parser.add_argument(
        '--type', required=True, help='The type of Kubernetes object, e.g., configmap. Required.')
    platform_parser.add_argument(
        '--url', help='The direct api path (metadata.selfLink) of the object, which overrides --type --name and --namespace options.')
    platform_parser.add_argument(
        '--regex', default=False, action="store_true", help='treat the --match value as a regex')
    platform_parser.add_argument(
        '--match-entity', help='the entity_check value to apply, i.e., "all", "at least one", "none exist"')
    platform_parser.add_argument(
        '--check-existence', help='check_existence` value for the `yamlfilecontent_test`.')
    platform_parser.add_argument(
        '--negate', default=False, action="store_true", help='negate the given matching criteria (does NOT match). Default is false.')
    platform_parser.set_defaults(func=createPlatformRuleFunc)

    node_parser = type_parser.add_parser('node', help='Creates a Node rule',  parents=[common_rule_args])
    node_parser.add_argument(
        '--template',  help='The tempate to use in a Node rule')
    node_parser.add_argument(
        '--template-vars',  help='The inputs for the template, coma separated')
    node_parser.set_defaults(func=createNodeRuleFunc)

    cluster_test_parser = subparser.add_parser(
        'cluster-test', help='Test a rule on a running OCP cluster using the compliance-operator.')
    cluster_test_parser.add_argument(
        '--rule', required=True, help='The name of the rule to test. Required.')
    cluster_test_parser.add_argument(
        '--skip-deploy', default=False, action="store_true", help='Skip deploying the compliance-operator. Default is to deploy.')
    cluster_test_parser.add_argument(
        '--skip-build', default=False, action="store_true", help='Skip building and pushing the data stream. Default is true.')
    cluster_test_parser.add_argument(
        '--scan-type', help='Type of scan to execute.', dest="scantype",
        default="Platform",
        choices=["Node", "Platform"])
    cluster_test_parser.add_argument(
        '--namespace', help='Namespace where compliance operator is installed. Default is "openshift-compliance".', dest="namespace", default="openshift-compliance"
    )
    cluster_test_parser.set_defaults(func=clusterTestFunc)

    test_parser = subparser.add_parser(
        'test', help='Test a rule locally against a directory of mocked object files using podman and an oscap container.')
    test_parser.add_argument('--rule', required=True,
                             help='The name of the rule to test.')
    test_parser.add_argument(
        '--contentdir', default="./build", help='The path to the directory containing the data stream')
    test_parser.add_argument(
        '--skip-build', default=False, action="store_true", help='Skip building the data stream. Default is false.')
    test_parser.add_argument('--objectdir', default="/tmp",
                             help='The path to a directory structure of yaml objects to test against.')
    test_parser.add_argument('--verbosity', default="INFO",
                             choices=['INFO', 'DEVEL'],
                             help='How verbose should OpenScap be')
    test_parser.set_defaults(func=testFunc)

    args = parser.parse_args()

    return args.func(args)


if __name__ == "__main__":
    sys.exit(main())



from __future__ import print_function

import collections
import contextlib
import fnmatch
import itertools
import json
import logging
import math
import os
import os.path
import re
import shutil
import tempfile

from ssg.constants import OSCAP_PROFILE, OSCAP_PROFILE_ALL_ID, OSCAP_RULE
from ssg.jinja import process_file_with_macros
from ssg.rules import is_rule_dir

from ssg_test_suite import oscap
from ssg_test_suite import xml_operations
from ssg_test_suite import test_env
from ssg_test_suite import common
from ssg_test_suite.log import LogHelper


Rule = collections.namedtuple(
    "Rule",
    ["directory", "id", "short_id", "template", "local_env_yaml", "rule"])

RuleTestContent = collections.namedtuple(
    "RuleTestContent", ["scenarios", "other_content"])

logging.getLogger(__name__).addHandler(logging.NullHandler())


def get_viable_profiles(selected_profiles, datastream, benchmark, script=None):
    """Read data stream, and return set intersection of profiles of given
    benchmark and those provided in `selected_profiles` parameter.
    """

    valid_profiles = []
    all_profiles_elements = xml_operations.get_all_profiles_in_benchmark(
        datastream, benchmark, logging)
    all_profiles = [el.attrib["id"] for el in all_profiles_elements]
    all_profiles.append(OSCAP_PROFILE_ALL_ID)

    for ds_profile in all_profiles:
        if 'ALL' in selected_profiles:
            valid_profiles += [ds_profile]
            continue
        for sel_profile in selected_profiles:
            if ds_profile.endswith(sel_profile):
                valid_profiles += [ds_profile]

    if not valid_profiles:
        if script:
            logging.warning('Script {0} - profile {1} not found in data stream'
                            .format(script, ", ".join(selected_profiles)))
        else:
            logging.warning('Profile {0} not found in data stream'
                            .format(", ".join(selected_profiles)))
    return valid_profiles


def generate_xslt_change_value_template(value_short_id, new_value):
    XSLT_TEMPLATE = """<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:ds="http://scap.nist.gov/schema/scap/source/1.2" xmlns:xccdf-1.2="http://checklists.nist.gov/xccdf/1.2">
    <xsl:output omit-xml-declaration="yes" indent="yes"/>
        <xsl:strip-space elements="*"/>
        <xsl:template match="node()|@*">
            <xsl:copy>
                <xsl:apply-templates select="node()|@*"/>
            </xsl:copy>
        </xsl:template>
        <xsl:template match="ds:component/xccdf-1.2:Benchmark//xccdf-1.2:Value[@id='xccdf_org.ssgproject.content_value_{value_short_id}']/xccdf-1.2:value[not(@selector)]/text()">{new_value}</xsl:template>
</xsl:stylesheet>"""
    return XSLT_TEMPLATE.format(value_short_id=value_short_id, new_value=new_value)


def _apply_script(rule_dir, test_env, script):
    """Run particular test script on VM and log it's output."""
    logging.debug("Applying script {0}".format(script))
    rule_name = os.path.basename(rule_dir)
    log_file_name = os.path.join(
        LogHelper.LOG_DIR, rule_name + ".prescripts.log")

    with open(log_file_name, 'a') as log_file:
        log_file.write('##### {0} / {1} #####\n'.format(rule_name, script))
        shared_dir = os.path.join(common.REMOTE_TEST_SCENARIOS_DIRECTORY, "shared")
        command = "cd {0}; SHARED={1} bash -x {2}".format(rule_dir, shared_dir, script)

        try:
            error_msg_template = (
                "Rule '{rule_name}' test setup script '{script}' "
                "failed with exit code {{rc}}".format(rule_name=rule_name, script=script)
            )
            test_env.execute_ssh_command(
                command, log_file, error_msg_template=error_msg_template)
        except RuntimeError as exc:
            logging.error(str(exc))
            return False
    return True


class RuleChecker(oscap.Checker):
    """
    Rule checks generally work like this -
    for every profile that supports that rule:

    - Alter the system.
    - Run the scan, check that the result meets expectations.
      If the test scenario passed as requested, return True,
      if it failed or passed unexpectedly, return False.

    The following sequence applies if the initial scan
    has failed as expected:

    - If there are no remediations, return True.
    - Run remediation, return False if it failed.
    - Return result of the final scan of remediated system.
    """
    def __init__(self, test_env):
        super(RuleChecker, self).__init__(test_env)

        self.results = list()
        self._current_result = None
        self.remote_dir = ""
        self.target_type = "rule ID"
        self.used_templated_test_scenarios = collections.defaultdict(set)
        self.rule_spec = None
        self.template_spec = None
        self.scenarios_profile = None

    def _run_test(self, profile, test_data):
        scenario = test_data["scenario"]
        rule_id = test_data["rule_id"]
        remediation_available = test_data["remediation_available"]

        LogHelper.preload_log(
            logging.INFO, "Script {0} using profile {1} OK".format(scenario.script, profile),
            log_target='pass')
        LogHelper.preload_log(
            logging.WARNING, "Script {0} using profile {1} notapplicable".format(scenario.script, profile),
            log_target='notapplicable')
        LogHelper.preload_log(
            logging.ERROR,
            "Script {0} using profile {1} found issue:".format(scenario.script, profile),
            log_target='fail')

        runner_cls = oscap.REMEDIATION_RULE_RUNNERS[self.remediate_using]
        runner_instance = runner_cls(
            self.test_env, oscap.process_profile_id(profile), self.datastream, self.benchmark_id,
            rule_id, scenario.script, self.dont_clean, self.no_reports, self.manual_debug)

        with runner_instance as runner:
            initial_scan_res = self._initial_scan_went_ok(runner, rule_id, scenario.context)
            if not initial_scan_res:
                return False
            if initial_scan_res == 2:
                # notapplicable
                return True

            supported_and_available_remediations = self._get_available_remediations(scenario)
            if (scenario.context not in ['fail', 'error']
                    or not supported_and_available_remediations):
                return True

            if remediation_available:
                if not self._remediation_went_ok(runner, rule_id):
                    return False

                return self._final_scan_went_ok(runner, rule_id)
            else:
                msg = ("No remediation is available for rule '{}'."
                       .format(rule_id))
                logging.warning(msg)
                return False

    def _initial_scan_went_ok(self, runner, rule_id, context):
        success = runner.run_stage_with_context("initial", context)
        self._current_result.record_stage_result("initial_scan", success)
        if not success:
            msg = ("The initial scan failed for rule '{}'."
                   .format(rule_id))
            logging.error(msg)
        return success

    def _is_remediation_available(self, rule):
        if xml_operations.find_fix_in_benchmark(
                self.datastream, self.benchmark_id, rule.id, self.remediate_using) is None:
            return False
        else:
            return True


    def _get_available_remediations(self, scenario):
        is_supported = set(['all'])
        is_supported.add(
            oscap.REMEDIATION_RUNNER_TO_REMEDIATION_MEANS[self.remediate_using])
        supported_and_available_remediations = set(
            scenario.script_params['remediation']).intersection(is_supported)
        return supported_and_available_remediations

    def _remediation_went_ok(self, runner, rule_id):
        success = runner.run_stage_with_context('remediation', 'fixed')
        self._current_result.record_stage_result("remediation", success)
        if not success:
            msg = ("The remediation failed for rule '{}'."
                   .format(rule_id))
            logging.error(msg)

        return success

    def _final_scan_went_ok(self, runner, rule_id):
        success = runner.run_stage_with_context('final', 'pass')
        self._current_result.record_stage_result("final_scan", success)
        if not success:
            msg = ("The check after remediation failed for rule '{}'."
                   .format(rule_id))
            logging.error(msg)
        return success


    def _rule_matches_rule_spec(self, rule_short_id):
        rule_id = OSCAP_RULE + rule_short_id
        if 'ALL' in self.rule_spec:
            return True
        else:
            for rule_to_be_tested in self.rule_spec:
                # we check for a substring
                if rule_to_be_tested.startswith(OSCAP_RULE):
                    pattern = rule_to_be_tested
                else:
                    pattern = OSCAP_RULE + rule_to_be_tested
                if fnmatch.fnmatch(rule_id, pattern):
                    return True
            return False

    def _rule_matches_template_spec(self, template):
        return True

    def _replace_platform_specific_packages(self, packages):
        """ Returns the provided package list with names
            updated according to the platform alternatives
            listed in platform_package_overrides product field """
        product_yaml = common.get_product_context(self.test_env.product)
        platform_package_overrides = product_yaml["platform_package_overrides"]
        packages_with_alternatives = set()
        for package in packages:
            if package in platform_package_overrides and platform_package_overrides[package]:
                packages_with_alternatives.add(platform_package_overrides[package])
            else:
                packages_with_alternatives.add(package)
        return packages_with_alternatives

    def _ensure_package_present_for_all_scenarios(
            self, test_content_by_rule_id):
        packages_required = set()

        for rule_test_content in test_content_by_rule_id.values():
            for s in rule_test_content.scenarios:
                scenario_packages = s.script_params["packages"]
                packages_required.update(scenario_packages)
        if packages_required:
            packages_to_install = self._replace_platform_specific_packages(packages_required)
            common.install_packages(self.test_env, packages_to_install)

    def _prepare_environment(self, test_content_by_rule_id):
        try:
            self.remote_dir = common.send_scripts(
                self.test_env, test_content_by_rule_id)
        except RuntimeError as exc:
            msg = "Unable to upload test scripts: {more_info}".format(more_info=str(exc))
            raise RuntimeError(msg)

        self._ensure_package_present_for_all_scenarios(test_content_by_rule_id)

    def _get_rules_to_test(self):
        """
        Returns:
            List of named tuples Rule having these fields:
                directory -- absolute path to the rule "tests" subdirectory
                            containing the test scenarios in Bash
                id -- full rule id as it is present in data stream
                short_id -- short rule ID, the same as basename of the directory
                            containing the test scenarios in Bash
                template -- name of the template the rule uses
                local_env_yaml -- env_yaml specific to rule's own context
                rule -- rule class, contains information parsed from rule.yml
        """

        # Here we need to perform some magic to handle parsing the rule (from a
        # product perspective) and loading any templated tests. In particular,
        # identifying which tests to potentially run involves invoking the
        # templating engine.
        #
        # Begin by loading context about our execution environment, if any.
        product = self.test_env.product
        product_yaml = common.get_product_context(product)
        all_rules_in_benchmark = xml_operations.get_all_rules_in_benchmark(
            self.datastream, self.benchmark_id)
        rules = []

        for dirpath, dirnames, filenames in common.walk_through_benchmark_dirs(
                product):
            if not is_rule_dir(dirpath):
                continue
            short_rule_id = os.path.basename(dirpath)
            full_rule_id = OSCAP_RULE + short_rule_id
            if not self._rule_matches_rule_spec(short_rule_id):
                continue
            if full_rule_id not in all_rules_in_benchmark:
                # This is a problem only if the user specified the rules to be
                # tested explicitly using command line arguments
                if self.target_type == "rule ID":
                    logging.warning(
                        "Rule '{0}' isn't present in benchmark '{1}' in '{2}'"
                        .format(
                            full_rule_id, self.benchmark_id, self.datastream))
                continue

            # Load the rule itself to check for a template.
            rule, local_env_yaml = common.load_rule_and_env(
                dirpath, product_yaml, product)

            tests_dir = os.path.join(dirpath, "tests")
            template_name = None
            if rule.template and rule.template['vars']:
                template_name = rule.template['name']
            if not self._rule_matches_template_spec(template_name):
                continue
            result = Rule(
                directory=tests_dir, id=full_rule_id,
                short_id=short_rule_id, template=template_name,
                local_env_yaml=local_env_yaml, rule=rule)
            rules.append(result)
        return rules

    def test_rule(self, state, rule, scenarios):
        remediation_available = self._is_remediation_available(rule)
        self._check_rule(
            rule, scenarios,
            self.remote_dir, state, remediation_available)

    def _slice_sbr(self, test_content_by_rule_id, slice_current, slice_total):
        """  Returns only a subset of test scenarios, representing slice_current-th
        slice out of slice_total"""

        tuple_repr = []
        for rule_id in test_content_by_rule_id:
            tuple_repr += itertools.product(
                [rule_id], test_content_by_rule_id[rule_id].scenarios)

        total_scenarios = len(tuple_repr)
        slice_low_bound = math.ceil(total_scenarios / slice_total * (slice_current - 1))
        slice_high_bound = math.ceil(total_scenarios / slice_total * slice_current)

        new_sbr = {}
        for rule_id, scenario in tuple_repr[slice_low_bound:slice_high_bound]:
            try:
                new_sbr[rule_id].scenarios.append(scenario)
            except KeyError:
                scenarios = [scenario]
                other_content = test_content_by_rule_id[rule_id].other_content
                new_sbr[rule_id] = RuleTestContent(scenarios, other_content)
        return new_sbr

    def _find_tests_paths(self, rule, product_yaml):
        # Start by checking for templating tests
        templated_tests_paths = common.fetch_templated_tests_paths(
            rule, product_yaml)

        # Add additional tests from the local rule directory. Note that,
        # like the behavior in template_tests, this will overwrite any
        # templated tests with the same file name.
        local_tests_paths = common.fetch_local_tests_paths(rule.directory)

        for filename in local_tests_paths:
            templated_tests_paths.pop(filename, None)
        if self.target_type != "template" and not self.test_env.duplicate_templates:
            for filename in self.used_templated_test_scenarios[rule.template]:
                templated_tests_paths.pop(filename, None)
            self.used_templated_test_scenarios[rule.template] |= set(
                templated_tests_paths.keys())
        return templated_tests_paths.values(), local_tests_paths.values()

    def _load_all_tests(self, rule):
        product_yaml = common.get_product_context(self.test_env.product)

        templated_tests_paths, local_tests_paths = self._find_tests_paths(
            rule, product_yaml)

        # All tests is a mapping from path (in the tarball) to contents
        # of the test case. This is necessary because later code (which
        # attempts to parse headers from the test case) don't have easy
        # access to templated content. By reading it and returning it
        # here, we can save later code from having to understand the
        # templating system.
        all_tests = dict()
        templated_tests = common.load_templated_tests(
            templated_tests_paths, rule.rule.template,
            rule.local_env_yaml)
        local_tests = common.load_local_tests(
            local_tests_paths, rule.local_env_yaml)
        all_tests.update(templated_tests)
        all_tests.update(local_tests)
        return all_tests

    def _get_rule_test_content(self, rule):
        checks = xml_operations.find_checks_in_rule(self.datastream, self.benchmark_id, rule.id)
        all_tests = self._load_all_tests(rule)
        scenarios = []
        other_content = dict()
        for file_name, file_content in all_tests.items():
            scenario_matches_regex = r'.*\.[^.]*\.sh$'
            if re.search(scenario_matches_regex, file_name):
                scenario = Scenario(file_name, file_content)
                scenario.override_profile(self.scenarios_profile)
                if scenario.matches_regex_and_check_and_platform(
                        self.scenarios_regex, checks, self.benchmark_cpes):
                    scenarios.append(scenario)
            else:
                other_content[file_name] = file_content
        return RuleTestContent(scenarios, other_content)

    def _get_shared_test_content(self):
        product_yaml = common.get_product_context(self.test_env.product)
        other_content = dict()
        for dirpath, _, filenames in os.walk(common._SHARED_DIR):
            for file_name in filenames:
                file_path = os.path.join(dirpath, file_name)
                rel_path = os.path.relpath(file_path, common._SHARED_DIR)
                try:
                    file_content = process_file_with_macros(file_path, product_yaml)
                except Exception as e:
                    logging.error("Error processing file {0}: {1}".format(file_path, str(e)))
                    continue
                other_content[rel_path] = file_content
        return RuleTestContent([], other_content)

    def _get_test_content_by_rule_id(self, rules_to_test):
        test_content_by_rule_id = dict()
        for rule in rules_to_test:
            rule_test_content = self._get_rule_test_content(rule)
            test_content_by_rule_id[rule.id] = rule_test_content
        sliced_test_content_by_rule_id = self._slice_sbr(
            test_content_by_rule_id, self.slice_current, self.slice_total)
        sliced_test_content_by_rule_id["shared"] = self._get_shared_test_content()
        return sliced_test_content_by_rule_id

    def _test_target(self):
        rules_to_test = self._get_rules_to_test()
        source = self.rule_spec
        if not self.rule_spec:
            source = self.template_spec
        if not rules_to_test:
            logging.error("No tests found matching the {0}(s) '{1}'".format(
                self.target_type,
                ", ".join(source)))
            return

        test_content_by_rule_id = self._get_test_content_by_rule_id(
            rules_to_test)

        self._prepare_environment(test_content_by_rule_id)

        with test_env.SavedState.create_from_environment(
                self.test_env, "tests_uploaded") as state:
            for rule in rules_to_test:
                try:
                    scenarios = test_content_by_rule_id[rule.id].scenarios
                    self.test_rule(state, rule, scenarios)
                except KeyError:
                    # rule is not processed in given slice
                    pass

    def _check_rule(self, rule, scenarios, remote_dir, state, remediation_available):
        remote_rule_dir = os.path.join(remote_dir, rule.short_id)
        logging.info(rule.id)

        logging.debug("Testing rule directory {0}".format(rule.directory))

        args_list = [
            (s, remote_rule_dir, rule.id, remediation_available) for s in scenarios
        ]
        state.map_on_top(self._check_and_record_rule_scenario, args_list)

    def _check_and_record_rule_scenario(self, scenario, remote_rule_dir, rule_id, remediation_available):
        self._current_result = common.RuleResult()

        self._current_result.conditions = common.Scenario_conditions(
            self.test_env.name, self.test_env.scanning_mode,
            self.remediate_using, self.datastream)
        self._current_result.scenario = common.Scenario_run(rule_id, scenario.script)
        self._current_result.when = self.test_timestamp_str

        with self.copy_of_datastream():
            self._check_rule_scenario(scenario, remote_rule_dir, rule_id, remediation_available)
        self.results.append(self._current_result.save_to_dict())

    @contextlib.contextmanager
    def copy_of_datastream(self, new_filename=None):
        prefixed_name = common.get_prefixed_name("ds_modified")
        old_filename = self.datastream
        if not new_filename:
            descriptor, new_filename = tempfile.mkstemp(prefix=prefixed_name, dir="/tmp")
        os.close(descriptor)
        shutil.copy(old_filename, new_filename)
        self.datastream = new_filename
        yield new_filename
        self.datastream = old_filename
        os.unlink(new_filename)

    def _change_variable_value(self, varname, value):
        descriptor, xslt_filename = tempfile.mkstemp(prefix="xslt-change-value", dir="/tmp")
        os.close(descriptor)
        template = generate_xslt_change_value_template(varname, value)
        with open(xslt_filename, "w") as fp:
            fp.write(template)
        descriptor, temp_datastream = tempfile.mkstemp(prefix="ds-temp", dir="/tmp")
        os.close(descriptor)
        log_file_name = os.path.join(LogHelper.LOG_DIR, "env-preparation.log")
        with open(log_file_name, "a") as log_file:
            result = common.run_with_stdout_logging(
                    "xsltproc", ("--output", temp_datastream, xslt_filename, self.datastream),
                    log_file)
            if result.returncode:
                msg = (
                    "Error changing value of '{varname}': {stderr}"
                    .format(varname=varname, stderr=result.stderr)
                )
                raise RuntimeError(msg)
        os.rename(temp_datastream, self.datastream)
        os.unlink(xslt_filename)

    def _verify_rule_presence(self, rule_id, script, profiles):
        for profile_id in profiles:
            if profile_id == OSCAP_PROFILE_ALL_ID:
                continue
            rules_in_profile = xml_operations.get_all_rule_ids_in_profile(
                self.datastream, self.benchmark_id, profile_id, logging)
            short_rule_id = rule_id.replace(OSCAP_RULE, "")
            if short_rule_id not in rules_in_profile:
                logging.warning(
                    "Rule {0} isn't part of profile {1} requested by "
                    "script {2}.".format(rule_id, profile_id, script)
                )
                return False
        return True


    def _check_rule_scenario(self, scenario, remote_rule_dir, rule_id, remediation_available):
        if "sce" in scenario.script_params["check"] and not self.test_env.sce_support:
            logging.warning(
                "Scenario {0} is used for SCE testing but OpenSCAP installed "
                "on the back end doesn't support SCE. To test SCE, please "
                "install the openscap-engine-sce package on the back end.".format(scenario.script)
            )
            return
        if not _apply_script(
                remote_rule_dir, self.test_env, scenario.script):
            logging.error("Environment failed to prepare, skipping test")
            self._current_result.record_stage_result("preparation", False)
            return

        if scenario.script_params["variables"]:
            for assignment in scenario.script_params["variables"]:
                varname, value = assignment.split("=", 1)
                self._change_variable_value(varname, value)
        self._current_result.record_stage_result("preparation", True)
        logging.debug('Using test script {0} with context {1}'
                      .format(scenario.script, scenario.context))

        profiles = get_viable_profiles(
            scenario.script_params['profiles'],
            self.datastream, self.benchmark_id, scenario.script)
        logging.debug("viable profiles are {0}".format(profiles))
        if not self._verify_rule_presence(rule_id, scenario.script, profiles):
            return
        test_data = dict(scenario=scenario,
                         rule_id=rule_id,
                         remediation_available=remediation_available)
        self.run_test_for_all_profiles(profiles, test_data)

        self.executed_tests += 1

    def finalize(self):
        super(RuleChecker, self).finalize()
        with open(os.path.join(LogHelper.LOG_DIR, "results.json"), "w") as f:
            json.dump(self.results, f)


class Scenario():
    def __init__(self, script, script_contents):
        self.script = script
        self.context = self._get_script_context()
        self.contents = script_contents
        self.script_params = self._parse_parameters()

    def _get_script_context(self):
        """Return context of the script."""
        result = re.search(r'.*\.([^.]*)\.[^.]*$', self.script)
        if result is None:
            return None
        return result.group(1)

    def _parse_parameters(self):
        """Parse parameters from script header"""
        params = {
            'profiles': [],
            'templates': [],
            'packages': [],
            'platform': ['multi_platform_all'],
            'check': ['any'],
            'remediation': ['all'],
            'variables': [],
        }

        for parameter in params:
            found = re.search(
                r'^# {0} = (.*)$'.format(parameter),
                self.contents, re.MULTILINE)
            if found is None:
                continue
            if parameter == "variables":
                variables = []
                for token in found.group(1).split(','):
                    token = token.strip()
                    if '=' in token:
                        variables.append(token)
                    else:
                        variables[-1] += "," + token
                params["variables"] = variables
            else:
                splitted = found.group(1).split(',')
                params[parameter] = [value.strip() for value in splitted]

        if not params["profiles"]:
            params["profiles"].append(OSCAP_PROFILE_ALL_ID)
            logging.debug(
                "Added the {0} profile to the list of available profiles "
                "for {1}"
                .format(OSCAP_PROFILE_ALL_ID, self.script))

        return params

    def override_profile(self, scenarios_profile):
        if scenarios_profile:
            self.script_params['profiles'] = [scenarios_profile]

    def matches_regex(self, scenarios_regex):
        if scenarios_regex is not None:
            scenarios_pattern = re.compile(scenarios_regex)
            if scenarios_pattern.match(self.script) is None:
                logging.debug(
                    "Skipping script %s - it did not match "
                    "--scenarios regex" % self.script
                )
                return False
        return True

    def matches_platform(self, benchmark_cpes):
        if self.context is None:
            return False
        if common.matches_platform(
                self.script_params["platform"], benchmark_cpes):
            return True
        else:
            logging.warning(
                "Script %s is not applicable on given platform" %
                self.script)
            return False

    def matches_check(self, rule_checks):
        if "any" in self.script_params["check"]:
            return True
        for check in self.script_params["check"]:
            if check in rule_checks:
                return True
        logging.warning(
            "Script %s is not applicable for %s check type" %
            (self.script, ", ".join(self.script_params['check'])))
        return False


    def matches_regex_and_check_and_platform(self, scenarios_regex, checks, benchmark_cpes):
        return (
            self.matches_regex(scenarios_regex)
            and self.matches_check(checks)
            and self.matches_platform(benchmark_cpes))


def perform_rule_check(options):
    checker = RuleChecker(options.test_env)

    checker.datastream = options.datastream
    checker.benchmark_id = options.benchmark_id
    checker.remediate_using = options.remediate_using
    checker.dont_clean = options.dont_clean
    checker.no_reports = options.no_reports
    checker.manual_debug = options.manual_debug
    checker.benchmark_cpes = options.benchmark_cpes
    checker.scenarios_regex = options.scenarios_regex
    checker.slice_current = options.slice_current
    checker.slice_total = options.slice_total
    checker.keep_snapshots = options.keep_snapshots
    checker.rule_spec = options.target
    checker.template_spec = None
    checker.scenarios_profile = options.scenarios_profile
    # check if target is a complete profile ID, if not prepend profile prefix
    if (checker.scenarios_profile is not None and
            not checker.scenarios_profile.startswith(OSCAP_PROFILE) and
            not oscap.is_virtual_oscap_profile(checker.scenarios_profile)):
        checker.scenarios_profile = OSCAP_PROFILE+options.scenarios_profile
    checker.test_target()



#!/usr/bin/python3
from __future__ import print_function

import argparse
import contextlib
import logging
import os
import os.path
import re
import sys
import tempfile
import textwrap
import time
from glob import glob

ssg_dir = os.path.join(os.path.dirname(__file__), "..")
sys.path.append(ssg_dir)

from ssg_test_suite.log import LogHelper
import ssg_test_suite.oscap
import ssg_test_suite.test_env
import ssg_test_suite.profile
import ssg_test_suite.rule
import ssg_test_suite.combined
import ssg_test_suite.template
from ssg_test_suite import xml_operations
from ssg.constants import DERIVATIVES_PRODUCT_MAPPING


def parse_args():
    parser = argparse.ArgumentParser()

    common_parser = argparse.ArgumentParser(add_help=False)
    common_parser.set_defaults(test_env=None)

    backends = common_parser.add_mutually_exclusive_group(required=True)

    backends.add_argument(
        "--docker", dest="docker", metavar="BASE_IMAGE",
        help="Use Docker test environment with this base image.")
    backends.add_argument(
        "--container", dest="container", metavar="BASE_IMAGE",
        help="Use container test environment with this base image.")

    backends.add_argument(
        "--libvirt", dest="libvirt", metavar=("HYPERVISOR", "DOMAIN"), nargs=2,
        help="libvirt hypervisor and domain name. When the leading URI driver protocol "
        "is omitted from the hypervisor, qemu:/// protocol is assumed. "
        "Example of a hypervisor domain name tuple: system ssg-test-suite")

    common_parser.add_argument(
        "--datastream", dest="datastream", metavar="DATASTREAM",
        help="Path to the Source data stream on this machine which is going to be tested. "
        "If not supplied, autodetection is attempted by looking into the build directory.")

    common_parser.add_argument(
        "--product", dest="product", metavar="PRODUCT", default=None,
        help="Product to interpret tests as being run under; autodetected from data stream "
        "if it follows the ssg-<product>-ds*.xml naming convention.")

    benchmarks = common_parser.add_mutually_exclusive_group()
    benchmarks.add_argument(
        "--xccdf-id", dest="xccdf_id", metavar="REF-ID", default=None,
        help="Reference ID related to benchmark to be used.")

    benchmarks.add_argument(
        "--xccdf-id-number", dest="xccdf_id_number", metavar="REF-ID-SELECT", type=int, default=0,
        help="Selection number of reference ID related to benchmark to be used.")

    common_parser.add_argument(
            "--add-platform",
            metavar="<CPE REGEX>",
            default=None,
            help="DEPRECATED: Use --remove-platforms instead; "
            "Find all CPEs that are present in local OpenSCAP's CPE dictionary "
            "that match the provided regex, "
            "and add them as platforms to all data stream benchmarks. "
            "If the regex doesn't match anything, it will be treated "
            "as a literal CPE, and added as a platform. "
            "For example, use 'cpe:/o:fedoraproject:fedora:30' or 'enterprise_linux'.")
    common_parser.add_argument(
        "--remove-fips-certified",
        action="store_true",
        help="Remove dependencies on rule installed_OS_is_FIPS_certified from "
        "all OVAL definitions that depend on it.")
    common_parser.add_argument(
            "--remove-platforms",
            default=False,
            action="store_true",
            help="Remove any platforms from the Benchmark XML elements and Profile "
            "XML elements, essentially making the content applicable to any platform. "
            "Although more low level platforms such as packages or container/machine "
            "CPE are still applicable.")
    common_parser.add_argument(
            "--make-applicable-in-containers",
            default=False,
            action="store_true",
            help="Removes platform constraints from rules "
            "to enable testing these rules on container backends.")
    common_parser.add_argument(
            "--remove-ocp4-only",
            default=False,
            action="store_true",
            help="Removes ocp4-only platform constraint from rules "
            "to enable testing these rules on ocp4 backends.")
    common_parser.add_argument("--loglevel",
                               dest="loglevel",
                               metavar="LOGLEVEL",
                               default="INFO",
                               help="Default level of console output")
    common_parser.add_argument("--logdir",
                               dest="logdir",
                               metavar="LOGDIR",
                               default=None,
                               help="Directory to which all output is saved")

    common_parser.add_argument(
        "--mode",
        dest="scanning_mode",
        default="online",
        choices=("online", "offline"),
        help="What type of check to use - either "
        "Online check done by running oscap inside the concerned system, or "
        "offline check that examines the filesystem from the host "
        "(either may require extended privileges).")

    common_parser.add_argument(
        "--remediate-using",
        dest="remediate_using",
        default="oscap",
        choices=ssg_test_suite.oscap.REMEDIATION_RULE_RUNNERS.keys(),
        help="What type of remediations to use - openscap online one, "
        "or remediation done by using remediation roles "
        "that are saved to disk beforehand.")

    common_parser.add_argument(
        "--duplicate-templates",
        dest="duplicate_templates",
        default=False,
        action="store_true",
        help="Execute all tests even for tests using shared templates; "
        "otherwise, executes one test per template type")

    common_parser.add_argument(
        "--keep-snapshots",
        dest="keep_snapshots",
        action="store_true",
        help="Do not remove existing snapshots created during previous tests.")

    subparsers = parser.add_subparsers(dest="subparser_name",
                                       help="Subcommands: profile, rule, combined")
    subparsers.required = True

    parser_profile = subparsers.add_parser("profile",
                                           formatter_class=argparse.RawDescriptionHelpFormatter,
                                           epilog=textwrap.dedent("""\
                    In case that tested profile contains rules which might prevent root ssh access
                    to the testing VM consider unselecting these rules. To unselect certain rules
                    from a data stream use `ds_unselect_rules.sh` script. List of such rules already
                    exists, see `unselect_rules_list` file.
                    Example usage:
                        ./ds_unselect_rules.sh ../build/ssg-fedora-ds.xml unselect_rules_list
                                           """),
                                           help=("Testing profile-based "
                                                 "remediation applied on already "
                                                 "installed machine"),
                                           parents=[common_parser])
    parser_profile.set_defaults(func=ssg_test_suite.profile.perform_profile_check)
    parser_profile.add_argument("target",
                                nargs="+",
                                metavar="DSPROFILE",
                                help=("Profiles to be tested, 'ALL' means every "
                                      "profile of particular benchmark will be "
                                      "evaluated."))

    parser_rule = subparsers.add_parser("rule",
                                        help=("Testing remediations of particular "
                                              "rule for various situations"),
                                        parents=[common_parser])
    parser_rule.set_defaults(func=ssg_test_suite.rule.perform_rule_check)
    parser_rule.add_argument(
        "target",
        nargs="+",
        metavar="RULE",
        help=(
            "Rule or rules to be tested. Special value 'ALL' means every "
            "rule-testing scenario will be evaluated. The SSG rule ID prefix "
            "is appended automatically if not provided. Wildcards to match "
            "multiple rules are accepted."
            )
        )
    parser_rule.add_argument("--debug",
                             dest="manual_debug",
                             action="store_true",
                             help=("If an error is encountered, all execution "
                                   "on the VM / container will pause to allow "
                                   "debugging."))
    parser_rule.add_argument("--dontclean",
                             dest="dont_clean",
                             action="store_true",
                             help="Do not remove html reports of successful runs")
    parser_rule.add_argument("--no-reports",
                             dest="no_reports",
                             action="store_true",
                             help="Do not run oscap with --report and --results options")
    parser_rule.add_argument("--scenarios",
                             dest="scenarios_regex",
                             default=None,
                             help="Regular expression matching test scenarios to run")
    parser_rule.add_argument("--profile",
                             dest="scenarios_profile",
                             default=None,
                             help="Override the profile used for test scenarios."
                                  " Variable selections will be done according "
                                  "to this profile.")
    parser_rule.add_argument("--slice",
                             dest='_slices',
                             # real dest is postprocessed later:
                             # 'slice_current' and 'slice_total'
                             metavar=('X', 'Y'),
                             default=[1, 1],
                             nargs=2,
                             type=int,
                             help=("Allows to run only Xth slice of Y in total, to enable "
                                   "stable parallelization of the bigger test sets."))

    parser_combined = subparsers.add_parser("combined",
                                            help=("Tests all rules in a profile evaluating them "
                                                  "against their test scenarios."),
                                            parents=[common_parser])
    parser_combined.set_defaults(func=ssg_test_suite.combined.perform_combined_check)
    parser_combined.add_argument("--dontclean",
                                 dest="dont_clean",
                                 action="store_true",
                                 help="Do not remove html reports of successful runs")
    parser_combined.add_argument("--no-reports",
                                 dest="no_reports",
                                 action="store_true",
                                 help="Do not run oscap with --report and --results options")
    parser_combined.add_argument("--scenarios",
                                 dest="scenarios_regex",
                                 default=None,
                                 help="Regular expression matching test scenarios to run")
    parser_combined.add_argument("--slice",
                                 dest='_slices',
                                 # real dest is postprocessed later:
                                 # 'slice_current' and 'slice_total'
                                 metavar=('X', 'Y'),
                                 default=[1, 1],
                                 nargs=2,
                                 type=int,
                                 help=("Allows to run only Xth slice of Y in total, to enable "
                                       "stable parallelization of the bigger test sets."))
    parser_combined.add_argument("target",
                                 nargs="+",
                                 metavar="TARGET",
                                 help=("Profiles whose rules are to be tested. Each rule selected "
                                       "in a profile will be evaluated against all its test "
                                       "scenarios."))

    parser_template = subparsers.add_parser("template",
                                            help=("Tests all rules in a template evaluating them "
                                                  "against their test scenarios."),
                                            parents=[common_parser])
    parser_template.set_defaults(func=ssg_test_suite.template.perform_template_check)
    parser_template.add_argument("--dontclean",
                                 dest="dont_clean",
                                 action="store_true",
                                 help="Do not remove html reports of successful runs")
    parser_template.add_argument("--no-reports",
                                 dest="no_reports",
                                 action="store_true",
                                 help="Do not run oscap with --report and --results options")
    parser_template.add_argument("--scenarios",
                                 dest="scenarios_regex",
                                 default=None,
                                 help="Regular expression matching test scenarios to run")
    parser_template.add_argument("--profile",
                                 dest="scenarios_profile",
                                 default=None,
                                 help="Override the profile used for test scenarios."
                                      " Variable selections will be done according "
                                      "to this profile.")
    parser_template.add_argument("--slice",
                                 dest='_slices',
                                 # real dest is postprocessed later:
                                 # 'slice_current' and 'slice_total'
                                 metavar=('X', 'Y'),
                                 default=[1, 1],
                                 nargs=2,
                                 type=int,
                                 help=("Allows to run only Xth slice of Y in total, to enable "
                                       "stable parallelization of the bigger test sets."))
    parser_template.add_argument("target",
                                 nargs="+",
                                 metavar="TARGET",
                                 help=("Template whose rules are to be tested. Each rule using"
                                       "a template will be evaluated against all its test "
                                       "scenarios."))

    options = parser.parse_args()
    if options.subparser_name in ["rule", "combined", "template"]:
        options.slice_current, options.slice_total = options._slices
        if options.slice_current < 1:
            raise argparse.ArgumentTypeError('Current slice needs to be positive integer')
        if options.slice_total < 1:
            raise argparse.ArgumentTypeError('Number of slices needs to be positive integer')
        if options.slice_current > options.slice_total:
            raise argparse.ArgumentTypeError(
                'Current slice cannot be greater than number of slices')
    return options


def get_logging_dir(options):
    body = 'custom'
    if 'ALL' in options.target:
        body = 'ALL'

    generic_logdir_stem = "{0}-{1}".format(options.subparser_name, body)

    if options.logdir is None:

        date_string = time.strftime('%Y-%m-%d-%H%M', time.localtime())
        logging_dir = os.path.join(
            os.getcwd(), 'logs', '{0}-{1}'.format(
                generic_logdir_stem, date_string))
        logging_dir = LogHelper.find_name(logging_dir)
    else:
        logging_dir = LogHelper.find_name(options.logdir)

    return logging_dir


def _print_available_benchmarks(xccdf_ids, n_xccdf_ids):
    logging.info("The data stream contains {0} Benchmarks".format(n_xccdf_ids))
    for i in range(0, n_xccdf_ids):
        logging.info("{0} - {1}".format(i, xccdf_ids[i]))


def auto_select_xccdf_id(datastream, bench_number):
    xccdf_ids = xml_operations.get_all_xccdf_ids_in_datastream(datastream)
    n_xccdf_ids = len(xccdf_ids)

    if n_xccdf_ids == 0:
        msg = ("The provided data stream doesn't contain any Benchmark")
        raise RuntimeError(msg)

    if bench_number < 0 or bench_number >= n_xccdf_ids:
        _print_available_benchmarks(xccdf_ids, n_xccdf_ids)
        logging.info("Selected Benchmark is {0}".format(bench_number))

        msg = ("Please select a valid Benchmark number")
        raise RuntimeError(msg)

    if n_xccdf_ids > 1:
        _print_available_benchmarks(xccdf_ids, n_xccdf_ids)
        logging.info("Selected Benchmark is {0}".format(bench_number))

        logging.info("To select a different Benchmark, "
                     "use --xccdf-id-number option.")

    return xccdf_ids[bench_number]


def get_datastreams():
    ds_glob = "ssg-*-ds.xml"
    build_dir_path = [os.path.dirname(__file__) or ".", "..", "build"]
    glob_pattern = os.path.sep.join(build_dir_path + [ds_glob])
    datastreams = [os.path.normpath(p) for p in glob(glob_pattern)]
    return datastreams


def get_unique_datastream():
    datastreams = get_datastreams()
    if len(datastreams) == 1:
        return datastreams[0]
    msg = ("Autodetection of the data stream file is possible only when there is "
           "a single one in the build dir, but")
    if not datastreams:
        raise RuntimeError(msg + " there is none.")
    raise RuntimeError(
        msg + " there are {0} of them. Use the --datastream option to select "
        "e.g. {1}".format(len(datastreams), datastreams))


def get_product_id(ds_filename):
    product_regex = re.compile(r'^.*ssg-([a-zA-Z0-9]*)-(ds|ds-1\.2)\.xml$')
    match = product_regex.match(ds_filename)
    if not match:
        msg = "Unable to detect product without explicit --product: "
        msg += "data stream {0} lacks product name".format(ds_filename)
        raise RuntimeError(msg)
    product = match.group(1)
    if product in DERIVATIVES_PRODUCT_MAPPING:
        product = DERIVATIVES_PRODUCT_MAPPING[product]
    return product


@contextlib.contextmanager
def datastream_in_stash(current_location):
    tfile = tempfile.NamedTemporaryFile(prefix="ssgts-ds-")

    tfile.write(open(current_location, "rb").read())
    tfile.flush()
    yield tfile.name


def normalize_passed_arguments(options):
    targets = []
    for target in options.target:
        if ',' in target:
            targets.extend(target.split(","))
        else:
            targets.append(target)
    options.target = targets

    if 'ALL' in options.target:
        options.target = ['ALL']

    if not options.datastream:
        options.datastream = get_unique_datastream()

    if not options.product and options.datastream:
        options.product = get_product_id(options.datastream)

    if options.xccdf_id is None:
        options.xccdf_id = auto_select_xccdf_id(options.datastream,
                                                options.xccdf_id_number)
    try:
        bench_id = xml_operations.infer_benchmark_id_from_component_ref_id(
            options.datastream, options.xccdf_id)
        options.benchmark_id = bench_id
    except RuntimeError as exc:
        msg = "Error inferring benchmark ID from component refId: {}".format(str(exc))
        raise RuntimeError(msg)

    if options.docker:
        options.test_env = ssg_test_suite.test_env.DockerTestEnv(
            options.scanning_mode, options.docker)
        logging.info(
            "The base image option has been specified, "
            "choosing Docker-based test environment.")
    elif options.container:
        options.test_env = ssg_test_suite.test_env.PodmanTestEnv(
            options.scanning_mode, options.container)
        logging.info(
            "The base image option has been specified, "
            "choosing Podman-based test environment.")
    else:
        hypervisor, domain_name = options.libvirt
        # Possible hypervisor spec we have to catch: qemu+unix:///session
        if not re.match(r"[\w\+]+:///", hypervisor):
            hypervisor = "qemu:///" + hypervisor
        options.test_env = ssg_test_suite.test_env.VMTestEnv(
            options.scanning_mode, hypervisor, domain_name, options.keep_snapshots)
        logging.info(
            "The base image option has not been specified, "
            "choosing libvirt-based test environment.")

    # Add in product to the test environment. This is independent of actual
    # test environment type so we do it after creation.
    options.test_env.product = options.product
    options.test_env.duplicate_templates = options.duplicate_templates

    try:
        benchmark_cpes = xml_operations.benchmark_get_applicable_platforms(
            options.datastream, options.benchmark_id
        )
        options.benchmark_cpes = benchmark_cpes
    except RuntimeError as exc:
        msg = "Error inferring platform from benchmark: {}".format(str(exc))
        raise RuntimeError(msg)


def main():
    options = parse_args()

    log = logging.getLogger()
    # this is general logger level - needs to be
    # debug otherwise it cuts silently everything
    log.setLevel(logging.DEBUG)

    LogHelper.add_console_logger(log, options.loglevel)

    try:
        normalize_passed_arguments(options)
    except RuntimeError as exc:
        msg = "Error occurred during options normalization: {}".format(str(exc))
        logging.error(msg)
        sys.exit(1)
    # logging dir needs to be created based on other options
    # thus we have to postprocess it

    logging_dir = get_logging_dir(options)

    LogHelper.add_logging_dir(log, logging_dir)

    with datastream_in_stash(options.datastream) as stashed_datastream:
        options.datastream = stashed_datastream

        with xml_operations.datastream_root(stashed_datastream, stashed_datastream) as root:
            if options.remove_platforms:
                xml_operations.remove_platforms(root)
            if options.make_applicable_in_containers:
                xml_operations.make_applicable_in_containers(root)
            if options.remove_ocp4_only:
                xml_operations.remove_ocp4_platforms(root)
            if options.add_platform:
                xml_operations.add_platform_to_benchmark(root, options.add_platform)
            if options.remove_fips_certified:
                xml_operations.remove_fips_certified(root)

        options.func(options)


if __name__ == "__main__":
    main()



import logging

from .general import OVALBaseObject
from .oval_definition_references import OVALDefinitionReference
from .oval_entities import (
    load_definition,
    load_object,
    load_state,
    load_test,
    load_variable,
)


class ExceptionDuplicateOVALEntity(Exception):
    pass


def _is_external_variable(component):
    return "external_variable" in component.tag


def _handle_existing_id(component, component_dict):
    # ID is identical, but OVAL entities are semantically different =>
    # report and error and exit with failure
    # Fixes: https://github.com/ComplianceAsCode/content/issues/1275
    if (
        component != component_dict[component.id_]
        and not _is_external_variable(component)
        and not _is_external_variable(component_dict[component.id_])
    ):
        # This is an error scenario - since by skipping second
        # implementation and using the first one for both references,
        # we might evaluate wrong requirement for the second entity
        # => report an error and exit with failure in that case
        # See
        #   https://github.com/ComplianceAsCode/content/issues/1275
        # for a reproducer and what could happen in this case
        raise ExceptionDuplicateOVALEntity(
            (
                "ERROR: it's not possible to use the same ID: {} for two semantically"
                " different OVAL entities:\nFirst entity:\n{}\nSecond entity:\n{}\n"
                "Use different ID for the second entity!!!\n"
            ).format(
                component.id_,
                str(component),
                str(component_dict[component.id_]),
            )
        )
    elif not _is_external_variable(component):
        # If OVAL entity is identical, but not external_variable, the
        # implementation should be rewritten each entity to be present
        # just once
        logging.info(
            (
                "OVAL ID {} is used multiple times and should represent "
                "the same elements.\nRewrite the OVAL checks. Place the identical IDs"
                " into their own definition and extend this definition by it."
            ).format(component.id_)
        )


def add_oval_component(component, component_dict):
    if component.id_ not in component_dict:
        component_dict[component.id_] = component
    else:
        _handle_existing_id(component, component_dict)


def _copy_component(destination, source_of_components):
    for component in source_of_components.values():
        add_oval_component(component, destination)


def _remove_keys_from_dict(dict_, to_remove):
    for k in to_remove:
        dict_.pop(k, None)


def _keep_keys_in_dict(dict_, to_keep):
    to_remove = [key for key in dict_ if key not in to_keep]
    _remove_keys_from_dict(dict_, to_remove)


def _save_referenced_vars(ref, entity):
    ref.save_variables(entity.get_variable_references())


def _save_definitions_references(ref, definition):
    if definition.criteria:
        ref.save_tests(definition.criteria.get_test_references())
        ref.save_definitions(definition.criteria.get_extend_definition_references())


def _save_test_references(ref, test):
    ref.save_object(test.object_ref)
    ref.save_states(test.state_refs)


def _save_object_references(ref, object_):
    _save_referenced_vars(ref, object_)
    ref.save_states(object_.get_state_references())
    ref.save_objects(object_.get_object_references())


def _save_variable_references(ref, variable):
    _save_referenced_vars(ref, variable)
    ref.save_objects(variable.get_object_references())


class OVALContainer(OVALBaseObject):
    def __init__(self):
        super(OVALContainer, self).__init__("")
        self.definitions = {}
        self.tests = {}
        self.objects = {}
        self.states = {}
        self.variables = {}
        self.MAP_COMPONENT_DICT = {
            "definitions": self.definitions,
            "tests": self.tests,
            "objects": self.objects,
            "states": self.states,
            "variables": self.variables,
        }

    def _call_function_for_every_component(self, _function, object_):
        _function(self.definitions, object_.definitions)
        _function(self.tests, object_.tests)
        _function(self.objects, object_.objects)
        _function(self.states, object_.states)
        _function(self.variables, object_.variables)

    def load_definition(self, oval_definition_xml_el):
        definition = load_definition(oval_definition_xml_el)
        add_oval_component(definition, self.definitions)

    def load_test(self, oval_test_xml_el):
        test = load_test(oval_test_xml_el)
        add_oval_component(test, self.tests)

    def load_object(self, oval_object_xml_el):
        object_ = load_object(oval_object_xml_el)
        add_oval_component(object_, self.objects)

    def load_state(self, oval_state_xml_element):
        state = load_state(oval_state_xml_element)
        add_oval_component(state, self.states)

    def load_variable(self, oval_variable_xml_element):
        variable = load_variable(oval_variable_xml_element)
        add_oval_component(variable, self.variables)

    def add_content_of_container(self, container):
        self._call_function_for_every_component(_copy_component, container)

    @staticmethod
    def _skip_if_is_none(value, component_id):
        raise NotImplementedError()

    def _process_component(self, ref, type_, function_save_refs):
        source = self.MAP_COMPONENT_DICT.get(type_)
        to_process, id_getter = ref.get_to_process_dict_and_id_getter(type_)
        while to_process:
            id_ = id_getter()
            entity = source.get(id_)
            if self._skip_if_is_none(entity, id_):
                continue
            function_save_refs(ref, entity)

    def _process_definition_references(self, ref):
        self._process_component(
            ref,
            "definitions",
            _save_definitions_references,
        )

    def _process_test_references(self, ref):
        self._process_component(
            ref,
            "tests",
            _save_test_references,
        )

    def _process_object_references(self, ref):
        self._process_component(
            ref,
            "objects",
            _save_object_references,
        )

    def _process_state_references(self, ref):
        self._process_component(
            ref,
            "states",
            _save_referenced_vars,
        )

    def _process_variable_references(self, ref):
        self._process_component(
            ref,
            "variables",
            _save_variable_references,
        )

    def _process_objects_states_variables_references(self, ref):
        while (
            ref.to_process_objects or ref.to_process_states or ref.to_process_variables
        ):
            self._process_object_references(ref)
            self._process_state_references(ref)
            self._process_variable_references(ref)

    def get_all_references_of_definition(self, definition_id):
        if definition_id not in self.definitions:
            raise ValueError(
                "ERROR: OVAL definition '{}' doesn't exist.".format(definition_id)
            )
        ref = OVALDefinitionReference(definition_id)
        self._process_definition_references(ref)
        self._process_test_references(ref)
        self._process_objects_states_variables_references(ref)
        return ref

    def keep_referenced_components(self, ref):
        self._call_function_for_every_component(_keep_keys_in_dict, ref)

    def _translate(self, translator, dict_, store_defname=False):
        for component_id in list(dict_.keys()):
            component = dict_[component_id]
            component.translate_id(translator, store_defname)
            translated_id = translator.generate_id(component.tag_name, component_id)
            del dict_[component_id]
            dict_[translated_id] = component

    def translate_id(self, translator, store_defname=False):
        for dict_ in self.MAP_COMPONENT_DICT.values():
            self._translate(translator, dict_, store_defname)



import os
import sys
import pytest
from argparse import Namespace
from utils.profile_tool import command_most_used_rules

DATA_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "ssg-module", "data")
)
DATA_STREAM_PATH = os.path.join(DATA_DIR, "simple_data_stream.xml")


def get_fake_args():
    return Namespace(
        subcommand="most-used-rules", BENCHMARKS=[str(DATA_STREAM_PATH)], format="plain"
    )


@pytest.mark.skipif(sys.version_info[0] < 3, reason="requires python3")
def test_command(capsys):
    command_most_used_rules(get_fake_args())
    captured = capsys.readouterr()
    assert "xccdf_com.example.www_rule_test-pass: 1" in captured.out



from __future__ import print_function

import logging
import os
import os.path
import sys

from ssg.utils import mkdir_p

logging.getLogger(__name__).addHandler(logging.NullHandler())


class LogHelper(object):
    """Provide focal point for logging. LOG_DIR is useful when output of script
    is saved into file. Log preloading is a way to log outcome before
    the output itself.
    """
    FORMATTER = logging.Formatter('%(levelname)s - %(message)s')
    INTERMEDIATE_LOGS = {'pass': [], 'fail': [], 'notapplicable': []}
    LOG_DIR = None
    LOG_FILE = None

    @staticmethod
    def find_name(original_path, suffix=""):
        """Find file name which is still not present in given directory

        Returns
        path -- original_path + number + suffix
        """
        log_number = 0
        # search for valid log directory, as we can run more than one
        # run per minute, first has only timestamp, rest has numbered suffix
        logging.debug("Searching for viable {0}{1}".format(original_path,
                                                           suffix))
        while True:
            if not log_number:
                log_number_suffix = ""
            else:
                log_number_suffix = "-{0}".format(log_number)

            path = original_path + log_number_suffix + suffix
            if os.path.exists(path):
                log_number += 1
            else:
                break
        logging.debug("Found {0}".format(path))
        return path

    @classmethod
    def add_console_logger(cls, logger, level):
        """Convenience function to set defaults for console logger"""
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(cls.FORMATTER)
        console_handler.setLevel(level)
        print('Setting console output to log level {0}'.format(level),
              file=sys.stderr)
        logger.addHandler(console_handler)

    @classmethod
    def add_logging_dir(cls, logger, _dirname):
        """Convenience function to set default logging into file.

        Also sets LOG_DIR and LOG_FILE
        """
        mkdir_p(_dirname)
        logfile = os.path.join(_dirname, 'test_suite.log')

        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(cls.FORMATTER)
        logger.info('Logging into {0}'.format(logfile))
        logger.addHandler(file_handler)
        cls.LOG_DIR = _dirname
        cls.LOG_FILE = logfile

    @classmethod
    def preload_log(cls, log_level, log_line, log_target=None):
        """Save log for later use. Fill named buffer `log_target`with the log
        line for later use.

        Special case:
        If `log_target` is default, i.e. None, all buffers will be filled with
        the same log line.
        """

        if log_target is None:
            # None means "All"
            for target in cls.INTERMEDIATE_LOGS:
                cls.INTERMEDIATE_LOGS[target] += [(log_level, log_line)]
        else:
            cls.INTERMEDIATE_LOGS[log_target] += [(log_level, log_line)]

    @classmethod
    def log_preloaded(cls, log_target):
        """Log messages preloaded in one of the named buffers. Wipe out all
        buffers afterwards.
        """
        for log_level, log_line in cls.INTERMEDIATE_LOGS[log_target]:
            logging.log(log_level, log_line)
        # cleanup
        for target in cls.INTERMEDIATE_LOGS:
            cls.INTERMEDIATE_LOGS[target] = []



import os

from ssg.constants import BENCHMARKS
from ssg.variables import (
    get_variable_files_in_folder,
    get_variable_files,
    get_variable_options,
    get_variable_property,
    get_variables_by_products,
    get_variables_from_profiles,
    get_variable_values,
)


def setup_test_files(base_dir, benchmark_dirs, create_txt_file=False):
    """
    Create benchmark directories and populate them with .var and optionally .txt files.

    Args:
        base_dir (Path): The base temporary directory.
        benchmark_dirs (list[str]): List of benchmark folder paths to create.
        create_txt_file (bool): Whether to create an additional .txt file in each benchmark.
    """
    # Ensures the shared/macros directory exists even if in this case the testing example does
    # not use Jinja2 macros.
    os.makedirs(base_dir / "shared" / "macros", exist_ok=True)

    for benchmark_dir in benchmark_dirs:
        path = base_dir / benchmark_dir
        os.makedirs(path, exist_ok=True)
        var_file = path / "test.var"
        var_file.write_text(
            "options:\n  default: value\n  option1: value1\n  option2: value2\n"
            "title: Test Title\ndescription: Test Description\n"
        )
        if create_txt_file:
            txt_file = path / "test.txt"
            txt_file.write_text("options:\n  option: value\n")


def test_get_variable_files_in_folder(tmp_path):
    content_dir = tmp_path / "content"
    benchmark_dirs = ["app", "app/rules"]
    setup_test_files(content_dir, benchmark_dirs, create_txt_file=True)
    result = get_variable_files_in_folder(str(content_dir), benchmark_dirs[0])
    assert len(result) == 2
    assert all(os.path.basename(file) == "test.var" for file in result)


def test_get_variable_files(tmp_path):
    content_dir = tmp_path / "content"
    BENCHMARKS.add(content_dir)
    benchmark_dirs = ["app", "app/rules"]
    setup_test_files(content_dir, benchmark_dirs)
    result = get_variable_files(str(content_dir))
    assert len(result) == 2
    assert all(os.path.basename(file) == "test.var" for file in result)


def test_get_variable_options(tmp_path):
    content_dir = tmp_path / "content"
    benchmark_dirs = ["app", "app/rules"]
    setup_test_files(content_dir, benchmark_dirs)
    result = get_variable_options(str(content_dir), "test")
    assert result == {"default": "value", "option1": "value1", "option2": "value2"}


def test_get_variables_by_products():
    products = ["rhel9"]
    content_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../.."))
    result = get_variables_by_products(str(content_dir), products)
    assert "var_selinux_policy_name" in result
    assert "rhel9" in result["var_selinux_policy_name"]


def test_get_variable_values(tmp_path):
    content_dir = tmp_path / "content"
    products = ["product1", "product2"]
    setup_test_files(content_dir, products)

    profiles_variables = {
        "test": {
            "product1": {"profile1": "option1"},
            "product2": {"profile2": "option2"},
        }
    }
    result = get_variable_values(str(content_dir), profiles_variables)
    assert result["test"]["product1"]["profile1"] == "value1"
    assert result["test"]["product2"]["profile2"] == "value2"


def test_get_variables_from_profiles():
    class MockProfile:
        def __init__(self, product_id, profile_id, variables):
            self.product_id = product_id
            self.profile_id = profile_id
            self.variables = variables

    profiles = [
        MockProfile("product1", "profile1", {"var1": "value1", "var2": "value2"}),
        MockProfile("product1", "profile2", {"var1": "value3"}),
        MockProfile("product2", "profile1", {"var2": "value4"}),
    ]

    expected_result = {
        "var1": {
            "product1": {
                "profile1": "value1",
                "profile2": "value3",
            }
        },
        "var2": {
            "product1": {
                "profile1": "value2",
            },
            "product2": {
                "profile1": "value4",
            }
        }
    }

    result = get_variables_from_profiles(profiles)
    assert result == expected_result


def test_get_variable_property(tmp_path):
    content_dir = tmp_path / "content"
    benchmark_dirs = ["app", "app/rules"]
    setup_test_files(content_dir, benchmark_dirs)

    result = get_variable_property(str(content_dir), "test", "title")
    assert result == "Test Title"

    # Test for a non-existent property
    result = get_variable_property(str(content_dir), "test", "non_existent_property")
    assert result == ""

    # Test for a non-existent variable
    result = get_variable_property(str(content_dir), "non_existent_variable", "property_name")
    assert result == ""



from __future__ import print_function

import functools
import logging
import os
import re
import shutil
import subprocess
import tarfile
import tempfile
import time
from collections import namedtuple

import ssg.yaml
from ssg.build_cpe import ProductCPEs
from ssg.build_yaml import Rule as RuleYAML
from ssg.constants import MULTI_PLATFORM_MAPPING
from ssg.constants import FULL_NAME_TO_PRODUCT_MAPPING
from ssg.constants import OSCAP_RULE
from ssg.jinja import process_file_with_macros
from ssg.products import product_yaml_path, load_product_yaml
from ssg.rules import get_rule_dir_yaml
from ssg.utils import mkdir_p
from ssg_test_suite.log import LogHelper

import ssg.templates


Scenario_run = namedtuple(
    "Scenario_run",
    ("rule_id", "script"))
Scenario_conditions = namedtuple(
    "Scenario_conditions",
    ("backend", "scanning_mode", "remediated_by", "datastream"))

SSG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))

_BENCHMARK_DIRS = [
        os.path.abspath(os.path.join(SSG_ROOT, 'linux_os', 'guide')),
        os.path.abspath(os.path.join(SSG_ROOT, 'applications')),
        ]

_SHARED_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../shared'))

_SHARED_TEMPLATES = os.path.abspath(os.path.join(SSG_ROOT, 'shared/templates'))

TEST_SUITE_NAME="ssgts"
TEST_SUITE_PREFIX = "_{}".format(TEST_SUITE_NAME)
REMOTE_USER = "root"
REMOTE_USER_HOME_DIRECTORY = "/root"
REMOTE_TEST_SCENARIOS_DIRECTORY = os.path.join(REMOTE_USER_HOME_DIRECTORY, TEST_SUITE_NAME)

try:
    SSH_ADDITIONAL_OPTS = tuple(os.environ.get('SSH_ADDITIONAL_OPTIONS').split())
except AttributeError:
    # If SSH_ADDITIONAL_OPTIONS is not defined set it to empty tuple.
    SSH_ADDITIONAL_OPTS = tuple()

SSH_ADDITIONAL_OPTS = (
    "-o", "StrictHostKeyChecking=no",
    "-o", "UserKnownHostsFile=/dev/null",
) + SSH_ADDITIONAL_OPTS

TESTS_CONFIG_NAME = "test_config.yml"


def walk_through_benchmark_dirs(product=None):
    directories = _BENCHMARK_DIRS
    if product is not None:
        yaml_path = product_yaml_path(SSG_ROOT, product)
        product_base = os.path.dirname(yaml_path)
        product_yaml = load_product_yaml(yaml_path)
        benchmark_root = os.path.join(product_base, product_yaml['benchmark_root'])
        directories = [os.path.abspath(benchmark_root)]

    for dirname in directories:
        for dirpath, dirnames, filenames in os.walk(dirname):
            yield dirpath, dirnames, filenames


class Stage(object):
    NONE = 0
    PREPARATION = 1
    INITIAL_SCAN = 2
    REMEDIATION = 3
    FINAL_SCAN = 4


@functools.total_ordering
class RuleResult(object):
    STAGE_STRINGS = {
        "preparation",
        "initial_scan",
        "remediation",
        "final_scan",
    }

    """
    Result of a test suite testing rule under a scenario.

    Supports ordering by success - the most successful run orders first.
    """
    def __init__(self, result_dict=None):
        self.scenario = Scenario_run("", "")
        self.conditions = Scenario_conditions("", "", "", "")
        self.when = ""
        self.passed_stages = dict()
        self.passed_stages_count = 0
        self.success = False

        if result_dict:
            self.load_from_dict(result_dict)

    def load_from_dict(self, data):
        self.scenario = Scenario_run(data["rule_id"], data["scenario_script"])
        self.conditions = Scenario_conditions(
            data["backend"], data["scanning_mode"],
            data["remediated_by"], data["datastream"])
        self.when = data["run_timestamp"]

        self.passed_stages = {key: data[key] for key in self.STAGE_STRINGS if key in data}
        self.passed_stages_count = sum(self.passed_stages.values())

        self.success = data.get("final_scan", False)
        if not self.success:
            self.success = (
                "remediation" not in data
                and data.get("initial_scan", False))

    def save_to_dict(self):
        data = dict()
        data["rule_id"] = self.scenario.rule_id
        data["scenario_script"] = self.scenario.script

        data["backend"] = self.conditions.backend
        data["scanning_mode"] = self.conditions.scanning_mode
        data["remediated_by"] = self.conditions.remediated_by
        data["datastream"] = self.conditions.datastream

        data["run_timestamp"] = self.when

        for stage_str, result in self.passed_stages.items():
            data[stage_str] = result

        return data

    def record_stage_result(self, stage, successful):
        assert stage in self.STAGE_STRINGS, (
            "Stage name {name} is invalid, choose one from {choices}"
            .format(name=stage, choices=", ".join(self.STAGE_STRINGS))
        )
        self.passed_stages[stage] = successful

    def relative_conditions_to(self, other):
        if self.conditions == other.conditions:
            return self.when, other.when
        else:
            return tuple(self.conditions), tuple(other.conditions)

    def __eq__(self, other):
        return (self.success == other.success
                and tuple(self.passed_stages) == tuple(other.passed_stages))

    def __lt__(self, other):
        return self.passed_stages_count > other.passed_stages_count


def run_cmd_local(command, verbose_path, env=None):
    command_string = ' '.join(command)
    logging.debug('Running {}'.format(command_string))
    returncode, output = _run_cmd(command, verbose_path, env)
    return returncode, output


def _run_cmd(command_list, verbose_path, env=None):
    returncode = 0
    output = b""
    try:
        with open(verbose_path, 'w') as verbose_file:
            output = subprocess.check_output(
                command_list, stderr=verbose_file, env=env)
    except subprocess.CalledProcessError as e:
        returncode = e.returncode
        output = e.output
    return returncode, output.decode('utf-8')


def _get_platform_cpes(platform):
    ssg_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    if platform.startswith("multi_platform_"):
        try:
            products = MULTI_PLATFORM_MAPPING[platform]
        except KeyError:
            logging.error(
                "Unknown multi_platform specifier: %s is not from %s"
                % (platform, ", ".join(MULTI_PLATFORM_MAPPING.keys())))
            raise ValueError
        platform_cpes = set()
        for p in products:
            product_yaml_path = os.path.join(ssg_root, "products", p, "product.yml")
            product_yaml = load_product_yaml(product_yaml_path)
            p_cpes = ProductCPEs()
            p_cpes.load_product_cpes(product_yaml)
            p_cpes.load_content_cpes(product_yaml)
            platform_cpes |= set(p_cpes.get_product_cpe_names())
        return platform_cpes
    else:
        # scenario platform is specified by a full product name
        try:
            product = FULL_NAME_TO_PRODUCT_MAPPING[platform]
        except KeyError:
            logging.error(
                "Unknown product name: %s is not from %s"
                % (platform, ", ".join(FULL_NAME_TO_PRODUCT_MAPPING.keys())))
            raise ValueError
        product_yaml_path = os.path.join(ssg_root, "products", product, "product.yml")
        product_yaml = load_product_yaml(product_yaml_path)
        product_cpes = ProductCPEs()
        product_cpes.load_product_cpes(product_yaml)
        product_cpes.load_content_cpes(product_yaml)
        platform_cpes = set(product_cpes.get_product_cpe_names())
        return platform_cpes


def matches_platform(scenario_platforms, benchmark_cpes):
    if "multi_platform_all" in scenario_platforms:
        return True
    scenario_cpes = set()
    for p in scenario_platforms:
        scenario_cpes |= _get_platform_cpes(p)
    return len(scenario_cpes & benchmark_cpes) > 0


def run_with_stdout_logging(command, args, log_file):
    log_file.write("{0} {1}\n".format(command, " ".join(args)))
    result = subprocess.run(
            (command,) + args, encoding="utf-8", stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, check=False)
    if result.stdout:
        log_file.write("STDOUT: ")
        log_file.write(result.stdout)
    if result.stderr:
        log_file.write("STDERR: ")
        log_file.write(result.stderr)
    return result


def _exclude_garbage(tarinfo):
    file_name = tarinfo.name
    if file_name.endswith('pyc'):
        return None
    if file_name.endswith('swp'):
        return None
    return tarinfo


def _make_file_root_owned(tarinfo):
    if tarinfo:
        tarinfo.uid = 0
        tarinfo.gid = 0
        # set permission to 775
        tarinfo.mode = 509
    return tarinfo


def get_product_context(product_id=None):
    """
    Returns a product YAML context if any product is specified. Hard-coded to
    assume a debug build.
    """
    # Load product's YAML file if present. This will allow us to parse
    # tests in the context of the product we're executing under.
    product_yaml = dict()
    if product_id:
        product = load_product_yaml(product_yaml_path(SSG_ROOT, product_id))
        product_properties_path = os.path.join(SSG_ROOT, "product_properties")
        product.read_properties_from_directory(product_properties_path)
        product_yaml.update(product)

    # We could run into a DocumentationNotComplete error when loading a
    # rule's YAML contents. However, because the test suite isn't executed
    # in the context of a particular build (though, ideally it would be
    # linked), we may not know exactly whether the top-level rule/profile
    # we're testing is actually completed. Thus, forcibly set the required
    # property to bypass this error.
    product_yaml['cmake_build_type'] = 'Debug'

    # Set the Jinja processing environment to Test Suite,
    # this allows Jinja macros to behave differently in a content build time and in a test time.
    product_yaml['SSG_TEST_SUITE_ENV'] = True

    return product_yaml


def load_rule_and_env(rule_dir_path, env_yaml, product=None):
    """
    Loads a rule and returns the combination of the RuleYAML class and
    the corresponding local environment for that rule.
    """

    # First build the path to the rule.yml file
    rule_path = get_rule_dir_yaml(rule_dir_path)

    # Load rule content in our environment. We use this to satisfy
    # some implied properties that might be used in the test suite.
    # Make sure we normalize to a specific product as well so that
    # when we load templated content it is correct.
    rule = RuleYAML.from_yaml(rule_path, env_yaml)
    rule.normalize(product)

    # Our local copy of env_yaml needs some properties from rule.yml
    # for completeness.
    local_env_yaml = dict()
    local_env_yaml.update(env_yaml)
    local_env_yaml['rule_id'] = rule.id_
    local_env_yaml['rule_title'] = rule.title
    local_env_yaml['products'] = {product}

    return rule, local_env_yaml


def write_rule_test_content_to_dir(rule_dir, test_content):
    for scenario in test_content.scenarios:
        scenario_file_path = os.path.join(rule_dir, scenario.script)
        with open(scenario_file_path, "w") as f:
            f.write(scenario.contents)
    for rel_file_path, file_content in test_content.other_content.items():
        if os.path.dirname(rel_file_path) != "":
            # file_path contains a directory, make sure it exists
            subdir = os.path.join(rule_dir, os.path.dirname(rel_file_path))
            if not os.path.exists(subdir):
                os.mkdir(subdir)
        file_path = os.path.join(rule_dir, rel_file_path)
        with open(file_path, "w") as f:
            f.write(file_content)
            # Ensure newline at the end of the file because
            # process_file_with_macros strips it off
            f.write("\n")


def create_tarball(test_content_by_rule_id):
    """
    Create a tarball which contains all test scenarios and additional
    content for every rule that is selected to be tested. The tarball contains
    directories with the test scenarios. The name of the directories is the
    same as short rule ID. There is no tree structure.
    """

    tmpdir = tempfile.mkdtemp()
    for rule_id, test_content in test_content_by_rule_id.items():
        short_rule_id = rule_id.replace(OSCAP_RULE, "")
        rule_dir = os.path.join(tmpdir, short_rule_id)
        mkdir_p(rule_dir)
        write_rule_test_content_to_dir(rule_dir, test_content)

    try:
        with tempfile.NamedTemporaryFile(
                "wb", suffix=".tar.gz", delete=False) as fp:
            with tarfile.TarFile.open(fileobj=fp, mode="w") as tarball:
                for rule_id in os.listdir(tmpdir):
                    # When a top-level directory exists under the temporary
                    # templated tests directory, we've already validated that
                    # it is a valid rule directory. Thus we can simply add it
                    # to the tarball.
                    absolute_dir = os.path.join(tmpdir, rule_id)
                    if not os.path.isdir(absolute_dir):
                        continue

                    tarball.add(
                        absolute_dir, arcname=rule_id,
                        filter=lambda tinfo: _exclude_garbage(_make_file_root_owned(tinfo))
                    )

            # Since we've added the templated contents into the tarball, we
            # can now delete the tree.
            shutil.rmtree(tmpdir, ignore_errors=True)
            return fp.name
    except Exception as exp:
        shutil.rmtree(tmpdir, ignore_errors=True)
        raise exp


def send_scripts(test_env, test_content_by_rule_id):
    remote_dir = REMOTE_TEST_SCENARIOS_DIRECTORY
    archive_file = create_tarball(test_content_by_rule_id)
    archive_file_basename = os.path.basename(archive_file)
    remote_archive_file = os.path.join(remote_dir, archive_file_basename)
    logging.debug("Uploading scripts.")
    log_file_name = os.path.join(LogHelper.LOG_DIR, "env-preparation.log")

    with open(log_file_name, 'a') as log_file:
        print("Setting up test setup scripts", file=log_file)

        test_env.execute_ssh_command(
            "mkdir -p {remote_dir}".format(remote_dir=remote_dir),
            log_file, "Cannot create directory {0}".format(remote_dir))
        test_env.scp_upload_file(
            archive_file, remote_dir,
            log_file, "Cannot copy archive {0} to the target machine's directory {1}"
            .format(archive_file, remote_dir))
        test_env.execute_ssh_command(
            "tar xf {remote_archive_file} -C {remote_dir}"
            .format(remote_dir=remote_dir, remote_archive_file=remote_archive_file),
            log_file, "Cannot extract data tarball {0}".format(remote_archive_file))
    os.unlink(archive_file)
    return remote_dir


def get_prefixed_name(state_name):
    return "{}_{}".format(TEST_SUITE_PREFIX, state_name)


def get_test_dir_config(test_dir, product_yaml):
    test_config = dict()
    test_config_filename = os.path.join(test_dir, TESTS_CONFIG_NAME)
    if os.path.exists(test_config_filename):
        test_config = ssg.yaml.open_and_expand(test_config_filename, product_yaml)
    return test_config


def select_templated_tests(test_dir_config, available_scenarios_basenames):
    deny_scenarios = set(test_dir_config.get("deny_templated_scenarios", []))
    available_scenarios_basenames = {
        test_name for test_name in available_scenarios_basenames
        if test_name not in deny_scenarios
    }

    allow_scenarios = set(test_dir_config.get("allow_templated_scenarios", []))
    if allow_scenarios:
        available_scenarios_basenames = {
            test_name for test_name in available_scenarios_basenames
            if test_name in allow_scenarios
        }

    allowed_and_denied = deny_scenarios.intersection(allow_scenarios)
    if allowed_and_denied:
        msg = (
            "Test directory configuration contain inconsistencies: {allowed_and_denied} "
            "scenarios are both allowed and denied."
            .format(allowed_and_denied=allowed_and_denied)
        )
        raise ValueError(msg)
    return available_scenarios_basenames


def fetch_templated_tests_paths(
        rule_namedtuple, product_yaml):
    rule = rule_namedtuple.rule
    if not rule.template or not rule.template['vars']:
        return dict()
    tests_paths = fetch_all_templated_tests_paths(rule.template)
    test_config = get_test_dir_config(rule_namedtuple.directory, product_yaml)
    allowed_tests_paths = select_templated_tests(
        test_config, tests_paths.keys())
    templated_test_scenarios = {
        name: tests_paths[name] for name in allowed_tests_paths}
    return templated_test_scenarios


def fetch_all_templated_tests_paths(rule_template):
    """
    Builds a dictionary of a test case relative path -> test case absolute path mapping.

    Here, we want to know what the relative path on disk (under the tests/
    subdirectory) is (such as "installed.pass.sh"), along with the actual
    absolute path.
    """
    template_name = rule_template['name']

    base_dir = os.path.abspath(os.path.join(_SHARED_TEMPLATES, template_name, "tests"))
    results = dict()

    # If no test cases exist, return an empty dictionary.
    if not os.path.exists(base_dir):
        return results

    # Walk files; note that we don't need to do anything about directories
    # as only files are recorded in the mapping; directories can be
    # inferred from the path.
    for dirpath, _, filenames in os.walk(base_dir):
        if not filenames:
            continue

        for filename in filenames:
            if filename.endswith(".swp"):
                continue

            # Relative path to the file becomes our results key.
            absolute_path = os.path.abspath(os.path.join(dirpath, filename))
            relative_path = os.path.relpath(absolute_path, base_dir)

            # Save the results under the relative path.
            results[relative_path] = absolute_path
    return results


def load_templated_tests(
        templated_tests_paths, template, local_env_yaml):
    templated_tests = dict()
    for path in templated_tests_paths:
        test = load_test(path, template, local_env_yaml)
        basename = os.path.basename(path)
        templated_tests[basename] = test
    return templated_tests


def load_test(absolute_path, rule_template, local_env_yaml):
    template_name = rule_template['name']
    template_vars = rule_template['vars']
    template_vars["_rule_id"] = local_env_yaml["rule_id"]
    # Load template parameters and apply it to the test case.
    maybe_template = ssg.templates.Template.load_template(_SHARED_TEMPLATES, template_name)
    if maybe_template is not None:
        template_parameters = maybe_template.preprocess(template_vars, "tests")
    else:
        raise ValueError("Rule uses template '{}' "
                         "which doesn't exist in '{}".format(template_name, _SHARED_TEMPLATES))

    jinja_dict = ssg.utils.merge_dicts(local_env_yaml, template_parameters)
    filled_template = ssg.jinja.process_file_with_macros(
        absolute_path, jinja_dict)
    return filled_template


def file_known_as_useless(file_name):
    return file_name.endswith(".swp")


def fetch_local_tests_paths(tests_dir):
    if not os.path.exists(tests_dir):
        return dict()
    all_tests = dict()
    tests_dir_files = os.listdir(tests_dir)
    for test_case in tests_dir_files:
        # Skip vim swap files, they are not relevant and cause Jinja
        # expansion tracebacks
        if file_known_as_useless(test_case):
            continue
        test_path = os.path.join(tests_dir, test_case)
        if os.path.isdir(test_path):
            continue
        all_tests[test_case] = test_path
    return all_tests


def load_local_tests(local_tests_paths, local_env_yaml):
    local_tests = dict()
    for path in local_tests_paths:
        test = process_file_with_macros(path, local_env_yaml)
        basename = os.path.basename(path)
        local_tests[basename] = test
    return local_tests


def get_cpe_of_tested_os(test_env, log_file):
    os_release_file = "/etc/os-release"
    cpe_line = test_env.execute_ssh_command(
        "grep CPE_NAME {os_release_file}".format(os_release_file=os_release_file),
        log_file)
    # We are parsing an assignment that is possibly quoted
    cpe = re.match(r'''CPE_NAME=(["']?)(.*)\1''', cpe_line)
    if cpe and cpe.groups()[1]:
        return cpe.groups()[1]
    msg = ["Unable to get a CPE of the system running tests"]
    if cpe_line:
        msg.append(
            "Retreived a CPE line that we couldn't parse: {cpe_line}"
            .format(cpe_line=cpe_line))
    else:
        msg.append(
            "Couldn't get CPE entry from '{os_release_file}'"
            .format(os_release_file=os_release_file))
    raise RuntimeError("\n".join(msg))


INSTALL_COMMANDS = dict(
    fedora=("dnf", "install", "-y"),
    ol7=("yum", "install", "-y"),
    ol8=("yum", "install", "-y"),
    ol9=("yum", "install", "-y"),
    rhel8=("yum", "install", "-y"),
    rhel9=("yum", "install", "-y"),
    rhel10=("dnf", "install", "-y"),
    sles=("zypper", "install", "-y"),
    ubuntu=("DEBIAN_FRONTEND=noninteractive", "apt", "install", "-y"),
    debian=("DEBIAN_FRONTEND=noninteractive", "apt", "install", "-y"),
)


def install_packages(test_env, packages):
    log_file_name = os.path.join(LogHelper.LOG_DIR, "env-preparation.log")

    with open(log_file_name, "a") as log_file:
        platform_cpe = get_cpe_of_tested_os(test_env, log_file)
    platform = cpes_to_platform([platform_cpe])

    command_str = " ".join(INSTALL_COMMANDS[platform] + tuple(packages))

    with open(log_file_name, 'a') as log_file:
        print("Installing packages", file=log_file)
        log_file.flush()
        test_env.execute_ssh_command(
            command_str, log_file,
            "Couldn't install required packages: {packages}".format(packages=",".join(packages)))


def _match_rhel_version(cpe):
    rhel_cpe = {
        "redhat:enterprise_linux": r":enterprise_linux:([^:]+):",
        "centos:centos": r"centos:centos:([0-9]+)"}
    for cpe_item in rhel_cpe.keys():
        if cpe_item in cpe:
            match = re.search(rhel_cpe.get(cpe_item), cpe)
            if match:
                major_version = match.groups()[0].split(".")[0]
                return "rhel" + major_version


def cpe_to_platform(cpe):
    trivials = ["fedora", "sles", "ubuntu", "debian"]
    for platform in trivials:
        if platform in cpe:
            return platform
    rhel_version = _match_rhel_version(cpe)
    if rhel_version is not None:
        return rhel_version
    if "oracle:linux" in cpe:
        match = re.search(r":linux:([^:]+):", cpe)
        if match:
            major_version = match.groups()[0]
            return "ol" + major_version


def cpes_to_platform(cpes):
    for cpe in cpes:
        platform = cpe_to_platform(cpe)
        if platform is not None:
            return platform
    msg = "Unable to deduce a platform from these CPEs: {cpes}".format(cpes=cpes)
    raise ValueError(msg)


def retry_with_stdout_logging(command, args, log_file, max_attempts=5):
    attempt = 0
    while attempt < max_attempts:
        result = run_with_stdout_logging(command, args, log_file)
        if result.returncode == 0:
            return result
        attempt += 1
        time.sleep(1)
    return result



import os
from typing import Optional

import ssg.products

from trestle.common.const import TRESTLE_GENERIC_NS
from trestle.core.generators import generate_sample_model
from trestle.oscal.common import Property

SSG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
VENDOR_ROOT = os.path.join(SSG_ROOT, "shared", "references", "oscal")
RULES_JSON = os.path.join(SSG_ROOT, "build", "rule_dirs.json")
BUILD_CONFIG = os.path.join(SSG_ROOT, "build", "build_config.yml")
TRESTLE_CD_NS = f"{TRESTLE_GENERIC_NS}/cd"
LOGGER_NAME = "oscal"


def get_benchmark_root(root: str, product: str) -> str:
    """Get the benchmark root."""
    product_yaml_path = ssg.products.product_yaml_path(root, product)
    product_yaml = ssg.products.load_product_yaml(product_yaml_path)
    product_dir = product_yaml.get("product_dir")
    benchmark_root = os.path.join(product_dir, product_yaml.get("benchmark_root"))
    return benchmark_root


def add_prop(name: str, value: str, remarks: Optional[str] = None) -> Property:
    """Add a property to a set of rule properties."""
    prop = generate_sample_model(Property)
    prop.name = name
    prop.value = value
    if remarks:
        prop.remarks = remarks
    prop.ns = TRESTLE_CD_NS  # type: ignore
    return prop



import pytest

import os
import ssg.playbook_builder
import yaml
import shutil


DATADIR = os.path.join(
    os.path.dirname(__file__),
    "test_playbook_builder_data"
)

build_config_yaml = os.path.join(DATADIR, "build_config.yml")
product_yaml = os.path.join(DATADIR, "product.yml")
input_dir = os.path.join(DATADIR, "fixes")
output_dir = os.path.join(DATADIR, "playbooks")
resolved_rules_dir = os.path.join(DATADIR, "rules")
resolved_profiles_dir = os.path.join(DATADIR, "profiles")
rule = "selinux_state"
profile = "ospp"

real_output_filepath = os.path.join(output_dir, profile, rule + ".yml")
expected_output_filepath = os.path.join(DATADIR, "selinux_state.yml")


def test_build_rule_playbook():
    playbook_builder = ssg.playbook_builder.PlaybookBuilder(
        product_yaml, input_dir, output_dir, resolved_rules_dir,
        resolved_profiles_dir, build_config_yaml
    )
    playbook_builder.build(profile, rule)

    assert os.path.exists(real_output_filepath)

    with open(real_output_filepath, "r") as real_output:
        real_output_yaml = yaml.load(real_output, Loader=yaml.Loader)
    with open(expected_output_filepath, "r") as expected_output:
        expected_output_yaml = yaml.load(expected_output, Loader=yaml.Loader)

    real_play = real_output_yaml.pop()
    expected_play = expected_output_yaml.pop()
    assert real_play["become"] == expected_play["become"]
    assert real_play["hosts"] == expected_play["hosts"]
    assert real_play["name"] == expected_play["name"]
    real_vars = real_play["vars"]
    expected_vars = expected_play["vars"]
    assert len(real_vars) == len(expected_vars)
    assert real_vars["var_selinux_state"] == expected_vars["var_selinux_state"]
    real_tags = real_play["tags"]
    expected_tags = expected_play["tags"]
    assert len(real_tags) == len(expected_tags)
    for tag in expected_tags:
        assert tag in real_tags
    for t1, t2 in zip(real_tags, expected_tags):
        assert t1 == t2
    real_tasks = real_play["tasks"]
    expected_tasks = expected_play["tasks"]
    assert len(real_tasks) == len(expected_tasks)
    real_task = real_tasks.pop()
    expected_task = expected_tasks.pop()
    assert real_task["name"] == expected_task["name"]
    assert len(real_task.keys()) == len(expected_task.keys())
    shutil.rmtree(output_dir)





import os
import pytest
import xml.etree.ElementTree as ET

import ssg.build_renumber
from ssg.constants import XCCDF12_NS, cce_uri, oval_namespace
from ssg.utils import SSGError

DATADIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


@pytest.fixture
def rule_el():
    rule = ET.Element("{%s}Rule" % XCCDF12_NS)
    rule.set("id", "xccdf_org.ssgproject_content_rule_selinux_state")
    return rule


@pytest.fixture
def rule_el2():
    rule = ET.Element("{%s}Rule" % XCCDF12_NS)
    rule.set("id", "xccdf_org.ssgproject_content_rule_accounts_tmout")
    return rule


@pytest.fixture
def rule_el_with_cce(rule_el):
    ident = ET.SubElement(rule_el, "{%s}ident" % XCCDF12_NS)
    ident.set("system", cce_uri)
    ident.text = "CCE-1234-789"
    return rule_el


@pytest.fixture
def rule_el2_with_cce(rule_el2):
    ident = ET.SubElement(rule_el2, "{%s}ident" % XCCDF12_NS)
    ident.set("system", cce_uri)
    ident.text = "CCE-666666"
    return rule_el2


def test_find_identcce(rule_el_with_cce):
    ident = ssg.build_renumber._find_identcce(rule_el_with_cce)
    assert ident is not None
    assert ident.get("system") == cce_uri
    assert ident.text == "CCE-1234-789"


@pytest.fixture
def rule_el_without_cce(rule_el):
    ident = ET.SubElement(rule_el, "{%s}ident" % XCCDF12_NS)
    ident.set("system", "http://seventy.com")
    ident.text = "70SS7070707070"
    return rule_el


def test_find_identcce_no_cce(rule_el_without_cce):
    ident = ssg.build_renumber._find_identcce(rule_el_without_cce)
    assert ident is None


@pytest.fixture
def benchmark_el():
    benchmark = ET.Element("{%s}Benchmark" % XCCDF12_NS)
    return benchmark


def test_rules_with_ids_generator_empty(benchmark_el):
    generator = ssg.build_renumber.rules_with_ids_generator(benchmark_el)
    assert list(generator) == []


@pytest.fixture
def benchmark_with_single_rule(benchmark_el, rule_el):
    benchmark_el.append(rule_el)
    return benchmark_el


@pytest.fixture
def benchmark_with_many_rules(benchmark_el, rule_el, rule_el2):
    benchmark_el.append(rule_el)
    benchmark_el.append(rule_el2)
    return benchmark_el


def test_rules_with_ids_generator_single_rule(benchmark_with_single_rule):
    generator = ssg.build_renumber.rules_with_ids_generator(
        benchmark_with_single_rule)
    items = list(generator)
    assert len(items) == 1
    id_, rule = items[0]
    assert id_ == "xccdf_org.ssgproject_content_rule_selinux_state"
    assert rule is not None
    assert rule.tag == "{%s}Rule" % XCCDF12_NS
    assert rule.get("id") == id_


def test_rules_with_ids_generator_(benchmark_with_many_rules):
    generator = ssg.build_renumber.rules_with_ids_generator(
        benchmark_with_many_rules)
    items = list(generator)
    assert len(items) == 2
    id_, rule = items[0]
    assert id_ == "xccdf_org.ssgproject_content_rule_selinux_state"
    assert rule is not None
    assert rule.tag == "{%s}Rule" % XCCDF12_NS
    assert rule.get("id") == id_
    id_, rule = items[1]
    assert id_ == "xccdf_org.ssgproject_content_rule_accounts_tmout"
    assert rule is not None
    assert rule.tag == "{%s}Rule" % XCCDF12_NS
    assert rule.get("id") == id_


@pytest.fixture
def benchmark_with_many_rules_cces(
        benchmark_el, rule_el_with_cce, rule_el2_with_cce):
    benchmark_el.append(rule_el_with_cce)
    benchmark_el.append(rule_el2_with_cce)
    return benchmark_el


def test_create_xccdf_id_to_cce_id_mapping(benchmark_with_many_rules_cces):
    mapping = ssg.build_renumber.create_xccdf_id_to_cce_id_mapping(
        benchmark_with_many_rules_cces)
    assert len(mapping) == 2
    assert mapping["xccdf_org.ssgproject_content_rule_selinux_state"] == \
        "CCE-1234-789"
    assert mapping["xccdf_org.ssgproject_content_rule_accounts_tmout"] == \
        "CCE-666666"


@pytest.fixture
def benchmark_with_many_rules_but_not_all_have_cces(
        benchmark_el, rule_el_without_cce, rule_el2_with_cce):
    benchmark_el.append(rule_el_without_cce)
    benchmark_el.append(rule_el2_with_cce)
    return benchmark_el


def test_create_xccdf_id_to_cce_id_mapping_2(
        benchmark_with_many_rules_but_not_all_have_cces):
    mapping = ssg.build_renumber.create_xccdf_id_to_cce_id_mapping(
        benchmark_with_many_rules_but_not_all_have_cces)
    assert len(mapping) == 1
    assert mapping["xccdf_org.ssgproject_content_rule_accounts_tmout"] == \
        "CCE-666666"


@pytest.fixture
def oval_def_apple():
    odef = ET.Element("{%s}definition" % oval_namespace)
    odef.set("id", "apple")
    return odef


@pytest.fixture
def oval_def_pear():
    odef = ET.Element("{%s}definition" % oval_namespace)
    odef.set("id", "pear")
    criteria = ET.SubElement(odef, "{%s}criteria" % oval_namespace)
    edef = ET.SubElement(criteria, "{%s}extend_definition" % oval_namespace)
    edef.set("definition_ref", "apple")
    return odef


@pytest.fixture
def xccdf_with_invalid_cce():
    bench = ET.Element("{%s}Benchmark" % XCCDF12_NS)
    rule = ET.SubElement(bench, "{%s}Rule" % XCCDF12_NS)
    rule.set("id", "errata_devel")
    ident = ET.SubElement(rule, "{%s}ident" % XCCDF12_NS)
    ident.set("system", cce_uri)
    ident.text = "RHEL7-CCE-TBD"
    return bench


def test_verify_correct_form_of_referenced_cce_identifiers(
        xccdf_with_invalid_cce):
    with pytest.raises(SSGError) as e:
        ssg.build_renumber.verify_correct_form_of_referenced_cce_identifiers(
            xccdf_with_invalid_cce)
    assert (
        "Warning: CCE 'RHEL7-CCE-TBD' is invalid for rule 'errata_devel'. "
        "Removing CCE...") in str(e)


@pytest.fixture
def xccdf_with_no_cce():
    bench = ET.Element("{%s}Benchmark" % XCCDF12_NS)
    rule = ET.SubElement(bench, "{%s}Rule" % XCCDF12_NS)
    rule.set("id", "errata_devel")
    return bench


def test_verify_correct_form_of_referenced_cce_identifiers_no_cce(
        xccdf_with_no_cce):
    try:
        ssg.build_renumber.verify_correct_form_of_referenced_cce_identifiers(
            xccdf_with_no_cce)
    except SSGError as e:
        assert False, "Raised SSGError: " + str(e)



import utils.srg_export.data


def get_heading(row: dict) -> str:
    if row['STIGID'] != '':
        return row['STIGID']
    else:
        return row['SRGID']


def get_content(header: str, row: dict) -> str:
    if row[header] is None:
        return ''
    else:
        content = row[header]
        content = content.replace('\n', '\n\n')
        content = content.replace('*', '&#42;')
        return content


def write_md_file(output, output_path):
    with open(output_path, 'w') as f:
        output_lines = '\n'.join(output)
        f.write(output_lines)


def handle_dict(data: list, output_path: str, title: str) -> None:
    """
    Given a dict with the fields for the srg export, create a formatted Markdown file

    To convert to a DOCX
    sudo dnf install pandoc texlive
    pip3 install git+https://github.com/pandocker/pandoc-docx-pagebreak-py
    pandoc -f markdown -t docx -o docx.docx --filter=pandoc-docx-pagebreakpy output.md

    """
    output = []
    for row in data:
        heading = get_heading(row)
        output.append(f'# {heading}')
        output.append('')
        for _, header in utils.srg_export.data.COLUMN_MAPPINGS.items():
            output.append(f'## {header}')
            content = get_content(header, row)
            output.append(content)
            output.append('')
        output.append('')
        output.append('\\newpage')
        output.append('')

    write_md_file(output, output_path)



from ssg.utils import parse_template_boolean_value
from ssg.utils import ensure_file_paths_and_file_regexes_are_correctly_defined

def map_symbolic_permissions(filemode, allow_stricter_permissions):
    mode_int = int(filemode, 8)
    fields = [
        ('o', 'x'), ('o', 'w'), ('o', 'r'),
        ('g', 'x'), ('g', 'w'), ('g', 'r'),
        ('u', 'x'), ('u', 'w'), ('u', 'r'),
        ('o', 't'), ('g', 's'), ('u', 's')
    ]
    mode_dict = {'u': '', 'g': '', 'o': ''}
    for field in fields:
        if mode_int & 0x01 == 1:
            if not allow_stricter_permissions:
                mode_dict[field[0]] += field[1]
        else:
            if allow_stricter_permissions:
                mode_dict[field[0]] += field[1]
        mode_int = mode_int >> 1
    return mode_dict


def group_symbolic_permissions(mode_dict):
    search_mode = ''
    fix_mode = ''
    for k in mode_dict:
        if mode_dict[k] != '':
            if search_mode != '':
                search_mode += ','
            search_mode += "{}+{}".format(k, mode_dict[k])

            if fix_mode != '':
                fix_mode += ','
            fix_mode += "{}-{}".format(k, mode_dict[k])
    return search_mode, fix_mode


def preprocess(data, lang):
    ensure_file_paths_and_file_regexes_are_correctly_defined(data)

    data["allow_stricter_permissions"] = parse_template_boolean_value(data, parameter="allow_stricter_permissions", default_value=True)

    data["recursive"] = parse_template_boolean_value(data,
                                                     parameter="recursive",
                                                     default_value=False)

    if lang == "oval":
        data["fileid"] = data["_rule_id"].replace("file_permissions", "")
        # build the state that describes our mode
        # mode_str maps to STATEMODE in the template
        mode = data["filemode"]
        fields = [
            'oexec', 'owrite', 'oread', 'gexec', 'gwrite', 'gread',
            'uexec', 'uwrite', 'uread', 'sticky', 'sgid', 'suid']
        mode_int = int(mode, 8)
        mode_str = ""
        for field in fields:
            if mode_int & 0x01 == 1:
                if not data['allow_stricter_permissions']:
                    mode_str = (
                        "<unix:" + field + " datatype=\"boolean\">true</unix:"
                        + field + ">\n" + mode_str)
            else:
                mode_str = (
                    "<unix:" + field + " datatype=\"boolean\">false</unix:"
                    + field + ">\n" + mode_str)
            mode_int = mode_int >> 1
        data["statemode"] = mode_str.rstrip("\n")

    if lang in ["bash", "ansible"]:

        mode_map = map_symbolic_permissions(data["filemode"], data['allow_stricter_permissions'])
        search_mode, fix_mode = group_symbolic_permissions(mode_map)

        data["search_mode"] = search_mode
        if data['allow_stricter_permissions']:
            # overwrite "filemode" with a subtractive symbolic mode
            # e.g., if filemode was "0755", now it is "u-s,g-ws,o-wt"
            # This allows files to keep the permission in which they are already stricter
            data["filemode"] = fix_mode

    return data



from __future__ import absolute_import
from __future__ import print_function

import os
import os.path
import json

from .build_yaml import Rule, DocumentationNotComplete
from .constants import (
    MULTI_PLATFORM_LIST, OSCAP_VALUE, datastream_namespace,
    xlink_namespace, XCCDF12_NS, SCE_SYSTEM
)
from .jinja import process_file_with_macros
from .rules import get_rule_dir_id, get_rule_dir_sces, find_rule_dirs_in_paths
from . import utils, products


def write_sce_file(sce_content, output_dir, filename):
    with open(os.path.join(output_dir, filename), 'w') as output_file:
        print(sce_content, file=output_file)


def load_sce_and_metadata(file_path, local_env_yaml):
    """
    For the given SCE audit file (file_path) under the specified environment
    (local_env_yaml), parse the file while expanding Jinja macros and read any
    metadata headers the file contains. Note that the last keyword of a
    specified type is the recorded one.

    Returns (audit_content, metadata).
    """

    raw_content = process_file_with_macros(file_path, local_env_yaml)
    return load_sce_and_metadata_parsed(raw_content)


def _process_raw_content_line(line, sce_content, metadata):
    found_metadata = False
    keywords = ['platform', 'check-import', 'check-export', 'complex-check', 'environment']
    for keyword in keywords:
        if not line.startswith('# ' + keyword + ' = '):
            continue
        found_metadata = True
        # Strip off the initial comment marker
        _, value = line[2:].split('=', maxsplit=1)
        metadata[keyword] = value.strip()

    if not found_metadata:
        sce_content.append(line)


def _parse_metadata(raw_content):
    metadata = dict()
    sce_content = []
    shebang = "#!/usr/bin/bash"
    for line in raw_content.split("\n"):
        if line.startswith("#!"):
            shebang = line
            continue
        _process_raw_content_line(line, sce_content, metadata)
    return shebang, "\n".join(sce_content), metadata


def _set_metadata_default_values(metadata):
    if 'check-export' in metadata:
        # Special case for the variables exposed to the SCE script: prepend
        # the OSCAP_VALUE prefix to reference the variable
        new_variables = []
        for value in metadata['check-export'].split(','):
            k, v = value.split('=', maxsplit=1)
            new_variables.append(k+'='+OSCAP_VALUE+v)
        metadata['check-export'] = new_variables

    if 'platform' in metadata:
        metadata['platform'] = metadata['platform'].split(',')
    else:
        metadata['platform'] = ['multi_platform_all']

    if "environment" not in metadata:
        metadata["environment"] = "any"
    environment_options = ["normal", "bootc", "any"]
    if metadata["environment"] not in environment_options:
        raise RuntimeError(
            "Wrong value of the 'environment' headers: '%s'. It needs to be "
            "one of %s" % (
                metadata["environment"], ", ".join(environment_options))
        )


def _modify_sce_with_environment(sce_content, environment):
    if environment == "any":
        return sce_content
    if environment == "bootc":
        condition = "(rpm -q --quiet bootc && [ -e /run/.containerenv ])"
    if environment == "normal":
        condition = "! (rpm -q --quiet bootc && [ -e /run/.containerenv ])"
    lines = list(sce_content.split("\n"))
    for i in range(len(lines)):
        if len(lines[i]) > 0:
            lines[i] = (4 * " ") + lines[i]
    lines.insert(0, "if " + condition + " ; then")
    lines.append("else")
    lines.append("    echo \"The SCE check can't run in this environment.\"")
    lines.append("    exit \"$XCCDF_RESULT_ERROR\"")
    lines.append("fi")
    return "\n".join(lines)


def load_sce_and_metadata_parsed(raw_content):
    shebang, sce_content, metadata = _parse_metadata(raw_content)
    _set_metadata_default_values(metadata)
    sce_content = _modify_sce_with_environment(sce_content, metadata["environment"])
    sce_content = shebang + "\n" + sce_content
    return sce_content, metadata


def _check_is_loaded(already_loaded, filename):
    # Right now this check doesn't look at metadata or anything
    # else. Eventually we might add versions to the entry or
    # something.
    return filename in already_loaded


# Retrieve the SCE checks and return a list of path to each check script.
def collect_sce_checks(datastreamtree):
    checklists = datastreamtree.find(
        ".//{%s}checklists" % datastream_namespace)
    checklists_component_ref = checklists.find(
        "{%s}component-ref" % datastream_namespace)
    # The component ID is the component-ref href without leading '#'
    checklist_component_id = checklists_component_ref.get('{%s}href' % xlink_namespace)[1:]

    checks_xpath = str.format(
        ".//{{{ds_ns}}}component[@id='{cid}']/"
        "{{{xccdf_ns}}}Benchmark//"
        "{{{xccdf_ns}}}Rule//"
        "{{{xccdf_ns}}}check[@system='{sce_sys}']/"
        "{{{xccdf_ns}}}check-content-ref",
        ds_ns=datastream_namespace,
        xccdf_ns=XCCDF12_NS,
        cid=checklist_component_id,
        sce_sys=SCE_SYSTEM
    )

    checks = datastreamtree.findall(checks_xpath)
    # Extract the file paths of the SCE checks
    return [check.get('href') for check in checks]


class SCEBuilder():
    def __init__(self, env_yaml, product_yaml, template_builder, output_dir):
        self.env_yaml = env_yaml
        self.product_yaml = product_yaml
        self.template_builder = template_builder
        self.output_dir = output_dir
        self.already_loaded = dict()

    def _build_static_sce_check(self, rule_id, path, local_env_yaml):
        # To be compatible with later checks, use the rule_id (i.e., the
        # value of _dir) to recreate the expected filename if this OVAL
        # was in a rule directory. However, note that unlike
        # build_oval.checks(...), we have to get this script's extension
        # first.
        _, ext = os.path.splitext(path)
        filename = "{0}{1}".format(rule_id, ext)

        sce_content, metadata = load_sce_and_metadata(path, local_env_yaml)
        metadata['filename'] = filename
        product = utils.required_key(self.env_yaml, "product")

        if not utils.is_applicable_for_product(",".join(metadata["platform"]), product):
            return
        if _check_is_loaded(self.already_loaded, rule_id):
            return

        write_sce_file(sce_content, self.output_dir, filename)

        self.already_loaded[rule_id] = metadata

    def _get_rule_sce_lang(self, rule):
        langs = self.template_builder.get_resolved_langs_to_generate(rule)
        for lang in langs:
            if lang.name == 'sce-bash':
                return lang
        return None

    def _build_templated_sce_check(self, rule):
        if not rule.template:
            return
        sce_lang = self._get_rule_sce_lang(rule)
        if sce_lang is None:
            return

        # Here we know the specified rule has a template and this
        # template actually generates (bash) SCE content. We
        # prioritize bespoke SCE content over templated content,
        # however, while we add this to our metadata, we do not
        # bother (yet!) with generating the SCE content. This is done
        # at a later time by build-scripts/build_templated_content.py.
        if _check_is_loaded(self.already_loaded, rule.id_):
            return

        # While we don't _write_ it, we still need to parse SCE
        # metadata from the templated content. Render it internally.
        raw_sce_content = self.template_builder.get_lang_contents_for_templatable(
            rule, sce_lang
        )

        ext = '.sh'
        filename = rule.id_ + ext

        # Load metadata and infer correct file name.
        sce_content, metadata = load_sce_and_metadata_parsed(raw_sce_content)
        metadata['filename'] = filename

        # Skip the check if it isn't applicable for this product.
        product = utils.required_key(self.env_yaml, "product")
        if not utils.is_applicable_for_product(",".join(metadata["platform"]), product):
            return

        write_sce_file(sce_content, self.output_dir, filename)

        # Finally, include it in our loaded content
        self.already_loaded[rule.id_] = metadata

    def _build_rule(self, rule_dir_path):
        local_env_yaml = dict()
        local_env_yaml.update(self.env_yaml)
        product = utils.required_key(self.env_yaml, "product")
        rule_id = get_rule_dir_id(rule_dir_path)

        rule_path = os.path.join(rule_dir_path, "rule.yml")
        try:
            rule = Rule.from_yaml(rule_path, self.env_yaml)
        except DocumentationNotComplete:
            # Happens on non-debug builds when a rule isn't yet completed. We
            # don't want to build the SCE check for this rule yet so skip it
            # and move on.
            return

        local_env_yaml['rule_id'] = rule.id_
        local_env_yaml['rule_title'] = rule.title
        local_env_yaml['products'] = {product}

        for _path in get_rule_dir_sces(rule_dir_path, product):
            self._build_static_sce_check(rule_id, _path, local_env_yaml)

        self._build_templated_sce_check(rule)

    def _dump_metadata(self):
        # Finally, write out our metadata to disk so that we can reference it in
        # later build stages (such as during building shorthand content).
        metadata_path = os.path.join(self.output_dir, 'metadata.json')
        json.dump(self.already_loaded, open(metadata_path, 'w'))

    def _find_content_dirs(self):
        # We maintain the same search structure as build_ovals.py even though we
        # don't currently have any content under shared/checks/sce.
        product_dir = self.product_yaml["product_dir"]
        relative_guide_dir = utils.required_key(self.env_yaml, "benchmark_root")
        guide_dir = os.path.abspath(os.path.join(product_dir, relative_guide_dir))
        additional_content_directories = self.env_yaml.get(
            "additional_content_directories", [])
        add_content_dirs = [
            os.path.abspath(os.path.join(product_dir, rd))
            for rd in additional_content_directories
        ]
        return ([guide_dir] + add_content_dirs)

    def build(self):
        """
        Walks the content repository and builds all SCE checks (and metadata entry)
        into the output directory.
        """
        content_dirs = self._find_content_dirs()
        for _dir_path in find_rule_dirs_in_paths(content_dirs):
            self._build_rule(_dir_path)
        self._dump_metadata()



#!/usr/bin/python3

import argparse
import os.path
import codecs
import ssg
import ssg.contributors


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dry-run', action="store_true",
                        help="If set the script will not output.")
    return parser.parse_args()


def main():
    args = parse_args()

    contributors_md, contributors_xml = ssg.contributors.generate()

    if args.dry_run:
        exit(0)

    root_dir = os.path.dirname(os.path.dirname(__file__))
    with codecs.open(os.path.join(root_dir, "Contributors.md"),
                     mode="w", encoding="utf-8") as f:
        f.write(contributors_md)

    with codecs.open(os.path.join(root_dir, "Contributors.xml"),
                     mode="w", encoding="utf-8") as f:
        f.write(contributors_xml)

    print("Don't forget to commit Contributors.md and Contributors.xml!")


if __name__ == "__main__":
    main()
